import contextlib
import json
import logging

from collections.abc import AsyncGenerator
from typing import Any, NoReturn

import httpx

from google.protobuf.json_format import MessageToDict, Parse, ParseDict

from a2a.client.client import ClientCallContext
from a2a.client.errors import A2AClientError
from a2a.client.transports.base import ClientTransport
from a2a.client.transports.http_helpers import (
    get_http_args,
    send_http_request,
    send_http_stream_request,
)
from a2a.compat.v0_3 import (
    a2a_v0_3_pb2,
    conversions,
    proto_utils,
)
from a2a.compat.v0_3 import (
    types as types_v03,
)
from a2a.compat.v0_3.extension_headers import add_legacy_extension_header
from a2a.types.a2a_pb2 import (
    AgentCard,
    CancelTaskRequest,
    DeleteTaskPushNotificationConfigRequest,
    GetExtendedAgentCardRequest,
    GetTaskPushNotificationConfigRequest,
    GetTaskRequest,
    ListTaskPushNotificationConfigsRequest,
    ListTaskPushNotificationConfigsResponse,
    ListTasksRequest,
    ListTasksResponse,
    SendMessageRequest,
    SendMessageResponse,
    StreamResponse,
    SubscribeToTaskRequest,
    Task,
    TaskPushNotificationConfig,
)
from a2a.utils.constants import PROTOCOL_VERSION_0_3, VERSION_HEADER
from a2a.utils.errors import JSON_RPC_ERROR_CODE_MAP, MethodNotFoundError
from a2a.utils.telemetry import SpanKind, trace_class


logger = logging.getLogger(__name__)

_A2A_ERROR_NAME_TO_CLS = {
    error_type.__name__: error_type for error_type in JSON_RPC_ERROR_CODE_MAP
}


@trace_class(kind=SpanKind.CLIENT)
class CompatRestTransport(ClientTransport):
    """A backward compatible REST transport for A2A v0.3."""

    def __init__(
        self,
        httpx_client: httpx.AsyncClient,
        agent_card: AgentCard | None,
        url: str,
        subscribe_method_override: str | None = None,
    ):
        """Initializes the CompatRestTransport."""
        self.url = url.removesuffix('/')
        self.httpx_client = httpx_client
        self.agent_card = agent_card
        self._subscribe_method_override = subscribe_method_override
        self._subscribe_auto_method_override = subscribe_method_override is None

    async def send_message(
        self,
        request: SendMessageRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> SendMessageResponse:
        """Sends a non-streaming message request to the agent."""
        req_v03 = conversions.to_compat_send_message_request(
            request, request_id=0
        )
        req_proto = a2a_v0_3_pb2.SendMessageRequest(
            request=proto_utils.ToProto.message(req_v03.params.message),
            configuration=proto_utils.ToProto.message_send_configuration(
                req_v03.params.configuration
            ),
            metadata=proto_utils.ToProto.metadata(req_v03.params.metadata),
        )

        response_data = await self._execute_request(
            'POST',
            '/v1/message:send',
            context=context,
            json=MessageToDict(req_proto, preserving_proto_field_name=True),
        )

        resp_proto = ParseDict(
            response_data,
            a2a_v0_3_pb2.SendMessageResponse(),
            ignore_unknown_fields=True,
        )
        which = resp_proto.WhichOneof('payload')
        if which == 'task':
            return SendMessageResponse(
                task=conversions.to_core_task(
                    proto_utils.FromProto.task(resp_proto.task)
                )
            )
        if which == 'msg':
            return SendMessageResponse(
                message=conversions.to_core_message(
                    proto_utils.FromProto.message(resp_proto.msg)
                )
            )
        return SendMessageResponse()

    async def send_message_streaming(
        self,
        request: SendMessageRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> AsyncGenerator[StreamResponse]:
        """Sends a streaming message request to the agent and yields responses as they arrive."""
        req_v03 = conversions.to_compat_send_message_request(
            request, request_id=0
        )
        req_proto = a2a_v0_3_pb2.SendMessageRequest(
            request=proto_utils.ToProto.message(req_v03.params.message),
            configuration=proto_utils.ToProto.message_send_configuration(
                req_v03.params.configuration
            ),
            metadata=proto_utils.ToProto.metadata(req_v03.params.metadata),
        )

        async for event in self._send_stream_request(
            'POST',
            '/v1/message:stream',
            context=context,
            json=MessageToDict(req_proto, preserving_proto_field_name=True),
        ):
            yield event

    async def get_task(
        self,
        request: GetTaskRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> Task:
        """Retrieves the current state and history of a specific task."""
        params = {}
        if request.HasField('history_length'):
            params['historyLength'] = request.history_length

        response_data = await self._execute_request(
            'GET',
            f'/v1/tasks/{request.id}',
            context=context,
            params=params,
        )
        resp_proto = ParseDict(
            response_data, a2a_v0_3_pb2.Task(), ignore_unknown_fields=True
        )
        return conversions.to_core_task(proto_utils.FromProto.task(resp_proto))

    async def list_tasks(
        self,
        request: ListTasksRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> ListTasksResponse:
        """Retrieves tasks for an agent."""
        raise NotImplementedError(
            'ListTasks is not supported in A2A v0.3 REST.'
        )

    async def cancel_task(
        self,
        request: CancelTaskRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> Task:
        """Requests the agent to cancel a specific task."""
        response_data = await self._execute_request(
            'POST',
            f'/v1/tasks/{request.id}:cancel',
            context=context,
        )
        resp_proto = ParseDict(
            response_data, a2a_v0_3_pb2.Task(), ignore_unknown_fields=True
        )
        return conversions.to_core_task(proto_utils.FromProto.task(resp_proto))

    async def create_task_push_notification_config(
        self,
        request: TaskPushNotificationConfig,
        *,
        context: ClientCallContext | None = None,
    ) -> TaskPushNotificationConfig:
        """Sets or updates the push notification configuration for a specific task."""
        req_v03 = (
            conversions.to_compat_create_task_push_notification_config_request(
                request, request_id=0
            )
        )
        req_proto = a2a_v0_3_pb2.CreateTaskPushNotificationConfigRequest(
            parent=f'tasks/{request.task_id}',
            config_id=req_v03.params.push_notification_config.id,
            config=proto_utils.ToProto.task_push_notification_config(
                req_v03.params
            ),
        )
        response_data = await self._execute_request(
            'POST',
            f'/v1/tasks/{request.task_id}/pushNotificationConfigs',
            context=context,
            json=MessageToDict(req_proto, preserving_proto_field_name=True),
        )
        resp_proto = ParseDict(
            response_data,
            a2a_v0_3_pb2.TaskPushNotificationConfig(),
            ignore_unknown_fields=True,
        )
        return conversions.to_core_task_push_notification_config(
            proto_utils.FromProto.task_push_notification_config(resp_proto)
        )

    async def get_task_push_notification_config(
        self,
        request: GetTaskPushNotificationConfigRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> TaskPushNotificationConfig:
        """Retrieves the push notification configuration for a specific task."""
        response_data = await self._execute_request(
            'GET',
            f'/v1/tasks/{request.task_id}/pushNotificationConfigs/{request.id}',
            context=context,
        )
        resp_proto = ParseDict(
            response_data,
            a2a_v0_3_pb2.TaskPushNotificationConfig(),
            ignore_unknown_fields=True,
        )
        return conversions.to_core_task_push_notification_config(
            proto_utils.FromProto.task_push_notification_config(resp_proto)
        )

    async def list_task_push_notification_configs(
        self,
        request: ListTaskPushNotificationConfigsRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> ListTaskPushNotificationConfigsResponse:
        """Lists push notification configurations for a specific task."""
        raise NotImplementedError(
            'list_task_push_notification_configs not supported in v0.3 REST'
        )

    async def delete_task_push_notification_config(
        self,
        request: DeleteTaskPushNotificationConfigRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> None:
        """Deletes the push notification configuration for a specific task."""
        raise NotImplementedError(
            'delete_task_push_notification_config not supported in v0.3 REST'
        )

    async def subscribe(
        self,
        request: SubscribeToTaskRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> AsyncGenerator[StreamResponse]:
        """Reconnects to get task updates.

        This method implements backward compatibility logic for the subscribe
        endpoint. It first attempts to use POST, which is the official method
        for A2A subscribe endpoint. If the server returns 405 Method Not Allowed,
        it falls back to GET and remembers this preference for future calls
        on this transport instance. If both fail with 405, it will default back
        to POST for next calls but will not retry again.
        """
        subscribe_method = self._subscribe_method_override or 'POST'
        try:
            async for event in self._send_stream_request(
                subscribe_method,
                f'/v1/tasks/{request.id}:subscribe',
                context=context,
            ):
                yield event
        except A2AClientError as e:
            # Check for 405 Method Not Allowed in the cause (httpx.HTTPStatusError)
            cause = e.__cause__
            if (
                isinstance(cause, httpx.HTTPStatusError)
                and cause.response.status_code == httpx.codes.METHOD_NOT_ALLOWED
            ):
                if self._subscribe_method_override:
                    if self._subscribe_auto_method_override:
                        self._subscribe_auto_method_override = False
                        self._subscribe_method_override = 'POST'
                    raise
                else:
                    self._subscribe_method_override = 'GET'
                    async for event in self.subscribe(request, context=context):
                        yield event
            else:
                raise

    async def get_extended_agent_card(
        self,
        request: GetExtendedAgentCardRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> AgentCard:
        """Retrieves the Extended AgentCard."""
        card = self.agent_card
        if card and not card.capabilities.extended_agent_card:
            return card

        response_data = await self._execute_request(
            'GET', '/v1/card', context=context
        )
        resp_proto = ParseDict(
            response_data, a2a_v0_3_pb2.AgentCard(), ignore_unknown_fields=True
        )
        card = conversions.to_core_agent_card(
            proto_utils.FromProto.agent_card(resp_proto)
        )
        self.agent_card = card
        return card

    async def close(self) -> None:
        """Closes the httpx client."""
        await self.httpx_client.aclose()

    def _handle_http_error(self, e: httpx.HTTPStatusError) -> NoReturn:
        """Handles HTTP status errors and raises the appropriate A2AError."""
        try:
            with contextlib.suppress(httpx.StreamClosed):
                e.response.read()

            try:
                error_data = e.response.json()
            except (json.JSONDecodeError, ValueError, httpx.ResponseNotRead):
                error_data = {}

            error_type = error_data.get('type')
            message = error_data.get('message', str(e))

            if isinstance(error_type, str):
                exception_cls = _A2A_ERROR_NAME_TO_CLS.get(error_type)
                if exception_cls:
                    raise exception_cls(message) from e
        except (json.JSONDecodeError, ValueError):
            pass

        status_code = e.response.status_code
        if status_code == httpx.codes.NOT_FOUND:
            raise MethodNotFoundError(
                f'Resource not found: {e.request.url}'
            ) from e

        raise A2AClientError(f'HTTP Error {status_code}: {e}') from e

    async def _send_stream_request(
        self,
        method: str,
        path: str,
        context: ClientCallContext | None = None,
        *,
        json: dict[str, Any] | None = None,
    ) -> AsyncGenerator[StreamResponse]:
        http_kwargs = get_http_args(context)
        http_kwargs.setdefault('headers', {})
        http_kwargs['headers'][VERSION_HEADER.lower()] = PROTOCOL_VERSION_0_3
        add_legacy_extension_header(http_kwargs['headers'])

        async for sse_data in send_http_stream_request(
            self.httpx_client,
            method,
            f'{self.url}{path}',
            self._handle_http_error,
            json=json,
            **http_kwargs,
        ):
            event_proto = a2a_v0_3_pb2.StreamResponse()
            Parse(sse_data, event_proto, ignore_unknown_fields=True)
            yield conversions.to_core_stream_response(
                types_v03.SendStreamingMessageSuccessResponse(
                    result=proto_utils.FromProto.stream_response(event_proto)
                )
            )

    async def _send_request(self, request: httpx.Request) -> dict[str, Any]:
        return await send_http_request(
            self.httpx_client, request, self._handle_http_error
        )

    async def _execute_request(
        self,
        method: str,
        path: str,
        context: ClientCallContext | None = None,
        *,
        json: dict[str, Any] | None = None,
        params: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        http_kwargs = get_http_args(context)
        http_kwargs.setdefault('headers', {})
        http_kwargs['headers'][VERSION_HEADER.lower()] = PROTOCOL_VERSION_0_3
        add_legacy_extension_header(http_kwargs['headers'])

        request = self.httpx_client.build_request(
            method,
            f'{self.url}{path}',
            json=json,
            params=params,
            **http_kwargs,
        )
        return await self._send_request(request)



import os

import sut_agent


try:
    import vertexai
except ImportError as e:
    raise ImportError(
        'VertexTaskStore requires vertexai. '
        'Install with: '
        "'pip install a2a-sdk[vertex]'"
    ) from e

from a2a.contrib.tasks.vertex_task_store import VertexTaskStore


def main() -> None:
    """Main entrypoint."""
    project = os.environ.get('VERTEX_PROJECT')
    location = os.environ.get('VERTEX_LOCATION')
    base_url = os.environ.get('VERTEX_BASE_URL')
    api_version = os.environ.get('VERTEX_API_VERSION')
    agent_engine_resource_id = os.environ.get('AGENT_ENGINE_RESOURCE_ID')

    if (
        not project
        or not location
        or not base_url
        or not api_version
        or not agent_engine_resource_id
    ):
        raise ValueError(
            'Environment variables VERTEX_PROJECT, VERTEX_LOCATION, '
            'VERTEX_BASE_URL, VERTEX_API_VERSION, and '
            'AGENT_ENGINE_RESOURCE_ID must be defined'
        )

    client = vertexai.Client(
        project=project,
        location=location,
        http_options={'base_url': base_url, 'api_version': api_version},
    )

    sut_agent.serve(
        VertexTaskStore(
            client=client,
            agent_engine_resource_id=agent_engine_resource_id,
        )
    )


if __name__ == '__main__':
    main()



import asyncio
import logging
import sys

from abc import ABC, abstractmethod
from types import TracebackType
from typing import Any, cast

from typing_extensions import Self


if sys.version_info >= (3, 13):
    from asyncio import Queue as AsyncQueue
    from asyncio import QueueShutDown

    def _create_async_queue(maxsize: int = 0) -> AsyncQueue[Any]:
        """Create a backwards-compatible queue object."""
        return AsyncQueue(maxsize=maxsize)
else:
    import culsans

    from culsans import AsyncQueue  # type: ignore[no-redef]
    from culsans import (
        AsyncQueueShutDown as QueueShutDown,  # type: ignore[no-redef]
    )

    def _create_async_queue(maxsize: int = 0) -> AsyncQueue[Any]:
        """Create a backwards-compatible queue object."""
        return culsans.Queue(maxsize=maxsize).async_q  # type: ignore[no-any-return]


from a2a.types.a2a_pb2 import (
    Message,
    Task,
    TaskArtifactUpdateEvent,
    TaskStatusUpdateEvent,
)
from a2a.utils.telemetry import SpanKind, trace_class


logger = logging.getLogger(__name__)


Event = Message | Task | TaskStatusUpdateEvent | TaskArtifactUpdateEvent
"""Type alias for events that can be enqueued."""

DEFAULT_MAX_QUEUE_SIZE = 1024


class EventQueue(ABC):
    """Base class and factory for EventQueueSource.

    EventQueue provides an abstraction for a queue of events that can be tapped
    by multiple consumers.
    EventQueue maintain main queue and source and maintain child queues in sync.
    GUARANTEE: All sinks (including the default one) will receive events in the exact same order.

    WARNING (Concurrency): All events from all sinks (both the default queue and any
    tapped child queues) must be regularly consumed and marked as done. If any single
    consumer stops processing and its queue reaches capacity, it can block the event
    dispatcher and stall the entire system, causing a widespread deadlock.

    WARNING (Memory Leak): Event queues spawn background tasks. To prevent memory
    and task leaks, all queue objects (both source and sinks) MUST be explicitly
    closed via `await queue.close()` or by using the async context manager (`async with queue:`).
    Child queues are automatically closed when parent queue is closed, but you
    should still close them explicitly to prevent queues from reaching capacity by
    unconsumed events.

    Typical usage:
    queue = EventQueue()
    child_queue1 = await queue.tap()
    child_queue2 = await queue.tap()

    async for event in child_queue1:
        do_some_work(event)
        child_queue1.task_done()
    """

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        """Redirects instantiation to EventQueueLegacy for backwards compatibility."""
        if cls is EventQueue:
            instance = EventQueueLegacy.__new__(EventQueueLegacy)
            EventQueueLegacy.__init__(instance, *args, **kwargs)
            return cast('Self', instance)
        return super().__new__(cls)

    @abstractmethod
    async def enqueue_event(self, event: Event) -> None:
        """Pushes an event into the queue.

        Only main queue can enqueue events. Child queues can only dequeue events.
        """


@trace_class(kind=SpanKind.SERVER)
class EventQueueLegacy(EventQueue):
    """Event queue for A2A responses from agent.

    Acts as a buffer between the agent's asynchronous execution and the
    server's response handling (e.g., streaming via SSE). Supports tapping
    to create child queues that receive the same events.
    """

    def __init__(self, max_queue_size: int = DEFAULT_MAX_QUEUE_SIZE) -> None:
        """Initializes the EventQueue."""
        # Make sure the `asyncio.Queue` is bounded.
        # If it's unbounded (maxsize=0), then `queue.put()` never needs to wait,
        # and so the streaming won't work correctly.
        if max_queue_size <= 0:
            raise ValueError('max_queue_size must be greater than 0')

        self._queue: AsyncQueue[Event] = _create_async_queue(
            maxsize=max_queue_size
        )
        self._children: list[EventQueueLegacy] = []
        self._is_closed = False
        self._lock = asyncio.Lock()
        logger.debug('EventQueue initialized.')

    @property
    def queue(self) -> AsyncQueue[Event]:
        """[DEPRECATED] Returns the underlying asyncio.Queue."""
        return self._queue

    async def __aenter__(self) -> Self:
        """Enters the async context manager, returning the queue itself."""
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exits the async context manager, ensuring close() is called."""
        await self.close()

    async def enqueue_event(self, event: Event) -> None:
        """Enqueues an event to this queue and all its children.

        Args:
            event: The event object to enqueue.
        """
        async with self._lock:
            if self._is_closed:
                logger.warning('Queue is closed. Event will not be enqueued.')
                return

        logger.debug('Enqueuing event of type: %s', type(event))

        try:
            await self.queue.put(event)
        except QueueShutDown:
            logger.warning('Queue was closed during enqueuing. Event dropped.')
            return

        for child in self._children:
            await child.enqueue_event(event)

    async def dequeue_event(self) -> Event:
        """Dequeues an event from the queue.

        This implementation expects that dequeue to raise an exception when
        the queue has been closed. In python 3.13+ this is naturally provided
        by the QueueShutDown exception generated when the queue has closed and
        the user is awaiting the queue.get method. Python<=3.12 this needs to
        manage this lifecycle itself. The current implementation can lead to
        blocking if the dequeue_event is called before the EventQueue has been
        closed but when there are no events on the queue. One way to avoid this
        is to use an async Task management solution to cancel the get task if the queue
        has closed or some other condition is met. The implementation of the
        EventConsumer uses an async.wait with a timeout to abort the
        dequeue_event call and retry, when it will return with a closed error.

        Returns:
            The next event from the queue.

        Raises:
            asyncio.QueueShutDown: If the queue has been closed and is empty.
        """
        async with self._lock:
            if self._is_closed and self.queue.empty():
                logger.warning('Queue is closed. Event will not be dequeued.')
                raise QueueShutDown('Queue is closed.')

        logger.debug('Attempting to dequeue event (waiting).')
        event = await self.queue.get()
        logger.debug('Dequeued event (waited) of type: %s', type(event))
        return event

    def task_done(self) -> None:
        """Signals that a formerly enqueued task is complete.

        Used in conjunction with `dequeue_event` to track processed items.
        """
        logger.debug('Marking task as done in EventQueue.')
        self.queue.task_done()

    async def tap(
        self, max_queue_size: int = DEFAULT_MAX_QUEUE_SIZE
    ) -> 'EventQueueLegacy':
        """Taps the event queue to create a new child queue that receives future events.

        Returns:
            A new `EventQueue` instance that will receive all events enqueued
            to this parent queue from this point forward.
        """
        logger.debug('Tapping EventQueue to create a child queue.')
        queue = EventQueueLegacy(max_queue_size=max_queue_size)
        self._children.append(queue)
        return queue

    async def close(self, immediate: bool = False) -> None:
        """Closes the queue for future push events and also closes all child queues.

        Args:
            immediate: If True, immediately flushes the queue, discarding all pending
                events, and causes any currently blocked `dequeue_event` calls to raise
                `QueueShutDown`. If False (default), the queue is marked as closed to new
                events, but existing events can still be dequeued and processed until the
                queue is fully drained.
        """
        logger.debug('Closing EventQueue.')
        async with self._lock:
            if self._is_closed and not immediate:
                return
            self._is_closed = True

        self.queue.shutdown(immediate)

        await asyncio.gather(
            *(child.close(immediate) for child in self._children)
        )
        if not immediate:
            await self.queue.join()

    def is_closed(self) -> bool:
        """Checks if the queue is closed."""
        return self._is_closed



from typing import Any

from a2a.helpers.proto_helpers import get_message_text
from a2a.server.context import ServerCallContext
from a2a.server.id_generator import (
    IDGenerator,
    IDGeneratorContext,
    UUIDGenerator,
)
from a2a.types.a2a_pb2 import (
    Message,
    SendMessageConfiguration,
    SendMessageRequest,
    Task,
)
from a2a.utils.errors import InvalidParamsError


class RequestContext:
    """Request Context.

    Holds information about the current request being processed by the server,
    including the incoming message, task and context identifiers, and related
    tasks.
    """

    def __init__(  # noqa: PLR0913
        self,
        call_context: ServerCallContext,
        request: SendMessageRequest | None = None,
        task_id: str | None = None,
        context_id: str | None = None,
        task: Task | None = None,
        related_tasks: list[Task] | None = None,
        task_id_generator: IDGenerator | None = None,
        context_id_generator: IDGenerator | None = None,
    ):
        """Initializes the RequestContext.

        Args:
            call_context: The server call context associated with this request.
            request: The incoming `SendMessageRequest` request payload.
            task_id: The ID of the task explicitly provided in the request or path.
            context_id: The ID of the context explicitly provided in the request or path.
            task: The existing `Task` object retrieved from the store, if any.
            related_tasks: A list of other tasks related to the current request (e.g., for tool use).
            task_id_generator: ID generator for new task IDs. Defaults to UUID generator.
            context_id_generator: ID generator for new context IDs. Defaults to UUID generator.
        """
        if related_tasks is None:
            related_tasks = []
        self._call_context = call_context
        self._params = request
        self._task_id = task_id
        self._context_id = context_id
        self._current_task = task
        self._related_tasks = related_tasks
        self._task_id_generator = (
            task_id_generator if task_id_generator else UUIDGenerator()
        )
        self._context_id_generator = (
            context_id_generator if context_id_generator else UUIDGenerator()
        )
        # If the task id and context id were provided, make sure they
        # match the request. Otherwise, create them
        if self._params:
            if task_id:
                self._params.message.task_id = task_id
                if task and task.id != task_id:
                    raise InvalidParamsError(message='bad task id')
            else:
                self._check_or_generate_task_id()
            if context_id:
                self._params.message.context_id = context_id
                if task and task.context_id != context_id:
                    raise InvalidParamsError(message='bad context id')
            else:
                self._check_or_generate_context_id()

    def get_user_input(self, delimiter: str = '\n') -> str:
        """Extracts text content from the user's message parts.

        Args:
            delimiter: The string to use when joining multiple text parts.

        Returns:
            A single string containing all text content from the user message,
            joined by the specified delimiter. Returns an empty string if no
            user message is present or if it contains no text parts.
        """
        if not self._params:
            return ''

        return get_message_text(self._params.message, delimiter)

    def attach_related_task(self, task: Task) -> None:
        """Attaches a related task to the context.

        This is useful for scenarios like tool execution where a new task
        might be spawned.

        Args:
            task: The `Task` object to attach.
        """
        self._related_tasks.append(task)

    @property
    def message(self) -> Message | None:
        """The incoming `Message` object from the request, if available."""
        return self._params.message if self._params else None

    @property
    def related_tasks(self) -> list[Task]:
        """A list of tasks related to the current request."""
        return self._related_tasks

    @property
    def current_task(self) -> Task | None:
        """The current `Task` object being processed."""
        return self._current_task

    @current_task.setter
    def current_task(self, task: Task | None) -> None:
        """Sets the current task object."""
        self._current_task = task

    @property
    def task_id(self) -> str | None:
        """The ID of the task associated with this context."""
        return self._task_id

    @property
    def context_id(self) -> str | None:
        """The ID of the conversation context associated with this task."""
        return self._context_id

    @property
    def configuration(self) -> SendMessageConfiguration | None:
        """The `SendMessageConfiguration` from the request, if available."""
        return self._params.configuration if self._params else None

    @property
    def call_context(self) -> ServerCallContext:
        """The server call context associated with this request."""
        return self._call_context

    @property
    def metadata(self) -> dict[str, Any]:
        """Metadata associated with the request, if available."""
        if self._params and self._params.metadata:
            return dict(self._params.metadata)
        return {}

    @property
    def tenant(self) -> str:
        """The tenant associated with this request."""
        return self._call_context.tenant

    @property
    def requested_extensions(self) -> set[str]:
        """Extensions that the client requested for this interaction."""
        return self._call_context.requested_extensions

    def _check_or_generate_task_id(self) -> None:
        """Ensures a task ID is present, generating one if necessary."""
        if not self._params:
            return

        if not self._task_id and not self._params.message.task_id:
            self._params.message.task_id = self._task_id_generator.generate(
                IDGeneratorContext(context_id=self._context_id)
            )
        if self._params.message.task_id:
            self._task_id = self._params.message.task_id

    def _check_or_generate_context_id(self) -> None:
        """Ensures a context ID is present, generating one if necessary."""
        if not self._params:
            return

        if not self._context_id and not self._params.message.context_id:
            self._params.message.context_id = (
                self._context_id_generator.generate(
                    IDGeneratorContext(task_id=self._task_id)
                )
            )
        if self._params.message.context_id:
            self._context_id = self._params.message.context_id



import asyncio
import logging

from a2a.server.context import ServerCallContext
from a2a.server.owner_resolver import OwnerResolver, resolve_user_scope
from a2a.server.tasks.push_notification_config_store import (
    PushNotificationConfigStore,
)
from a2a.types.a2a_pb2 import TaskPushNotificationConfig


logger = logging.getLogger(__name__)


class InMemoryPushNotificationConfigStore(PushNotificationConfigStore):
    """In-memory implementation of PushNotificationConfigStore interface.

    Stores push notification configurations in a nested dictionary in memory,
    keyed by owner then task_id.
    """

    def __init__(
        self,
        owner_resolver: OwnerResolver = resolve_user_scope,
    ) -> None:
        """Initializes the InMemoryPushNotificationConfigStore."""
        self.lock = asyncio.Lock()
        self._push_notification_infos: dict[
            str, dict[str, list[TaskPushNotificationConfig]]
        ] = {}
        self.owner_resolver = owner_resolver

    def _get_owner_push_notification_infos(
        self, owner: str
    ) -> dict[str, list[TaskPushNotificationConfig]]:
        return self._push_notification_infos.get(owner, {})

    async def set_info(
        self,
        task_id: str,
        notification_config: TaskPushNotificationConfig,
        context: ServerCallContext,
    ) -> None:
        """Sets or updates the push notification configuration for a task in memory."""
        owner = self.owner_resolver(context)
        if owner not in self._push_notification_infos:
            self._push_notification_infos[owner] = {}
        async with self.lock:
            owner_infos = self._push_notification_infos[owner]
            if task_id not in owner_infos:
                owner_infos[task_id] = []

            if not notification_config.id:
                notification_config.id = task_id

            # Remove existing config with the same ID
            for config in owner_infos[task_id]:
                if config.id == notification_config.id:
                    owner_infos[task_id].remove(config)
                    break

            owner_infos[task_id].append(notification_config)
            logger.debug(
                'Push notification config for task %s with config id %s for owner %s saved/updated.',
                task_id,
                notification_config.id,
                owner,
            )

    async def get_info(
        self,
        task_id: str,
        context: ServerCallContext,
    ) -> list[TaskPushNotificationConfig]:
        """Retrieves all push notification configurations for a task from memory, for the given owner.

        Used by the user-callable read endpoints.
        """
        owner = self.owner_resolver(context)
        async with self.lock:
            owner_infos = self._get_owner_push_notification_infos(owner)
            return list(owner_infos.get(task_id, []))

    async def get_info_for_dispatch(
        self,
        task_id: str,
    ) -> list[TaskPushNotificationConfig]:
        """Retrieves all push notification configurations for a task across all owners.

        Used by the push-notification dispatch path.
        """
        async with self.lock:
            results: list[TaskPushNotificationConfig] = []
            for all_configs in self._push_notification_infos.values():
                results.extend(all_configs.get(task_id, []))
            return results

    async def delete_info(
        self,
        task_id: str,
        context: ServerCallContext,
        config_id: str | None = None,
    ) -> None:
        """Deletes push notification configurations for a task from memory.

        If config_id is provided, only that specific configuration is deleted.
        If config_id is None, all configurations for the task for the owner are deleted.
        """
        owner = self.owner_resolver(context)
        async with self.lock:
            owner_infos = self._get_owner_push_notification_infos(owner)
            if task_id not in owner_infos:
                logger.warning(
                    'Attempted to delete push notification config for task %s, owner %s that does not exist.',
                    task_id,
                    owner,
                )
                return

            if config_id is None:
                del owner_infos[task_id]
                logger.info(
                    'Deleted all push notification configs for task %s, owner %s.',
                    task_id,
                    owner,
                )
            else:
                configurations = owner_infos[task_id]
                found = False
                for config in configurations:
                    if config.id == config_id:
                        configurations.remove(config)
                        found = True
                        break
                if found:
                    logger.info(
                        'Deleted push notification config %s for task %s, owner %s.',
                        config_id,
                        task_id,
                        owner,
                    )
                    if len(configurations) == 0:
                        del owner_infos[task_id]
                else:
                    logger.warning(
                        'Attempted to delete push notification config %s for task %s, owner %s that does not exist.',
                        config_id,
                        task_id,
                        owner,
                    )

            if not owner_infos:
                del self._push_notification_infos[owner]



from collections.abc import Awaitable, Callable
from typing import TYPE_CHECKING, Any


if TYPE_CHECKING:
    from starlette.requests import Request
    from starlette.responses import JSONResponse, Response
    from starlette.routing import Route

    _package_starlette_installed = True
else:
    try:
        from starlette.requests import Request
        from starlette.responses import JSONResponse, Response
        from starlette.routing import Route

        _package_starlette_installed = True
    except ImportError:
        Route = Any
        Request = Any
        Response = Any
        JSONResponse = Any

        _package_starlette_installed = False

from a2a.server.request_handlers.response_helpers import agent_card_to_dict
from a2a.types.a2a_pb2 import AgentCard
from a2a.utils.constants import AGENT_CARD_WELL_KNOWN_PATH


def create_agent_card_routes(
    agent_card: AgentCard,
    card_modifier: Callable[[AgentCard], Awaitable[AgentCard]] | None = None,
    card_url: str = AGENT_CARD_WELL_KNOWN_PATH,
) -> list['Route']:
    """Creates the Starlette Route for the A2A protocol agent card endpoint."""
    if not _package_starlette_installed:
        raise ImportError(
            'The `starlette` package is required to use `create_agent_card_routes`. '
            'It can be installed as part of `a2a-sdk` optional dependencies, `a2a-sdk[http-server]`.'
        )

    async def _get_agent_card(request: Request) -> Response:
        card_to_serve = agent_card
        if card_modifier:
            card_to_serve = await card_modifier(card_to_serve)
        return JSONResponse(agent_card_to_dict(card_to_serve))

    return [
        Route(
            path=card_url,
            endpoint=_get_agent_card,
            methods=['GET'],
        )
    ]



from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel


def to_camel_custom(snake: str) -> str:
    """Convert a snake_case string to camelCase.

    Args:
        snake: The string to convert.

    Returns:
        The converted camelCase string.
    """
    # First, remove any trailing underscores. This is common for names that
    # conflict with Python keywords, like 'in_' or 'from_'.
    if snake.endswith('_'):
        snake = snake.rstrip('_')
    return to_camel(snake)


class A2ABaseModel(BaseModel):
    """Base class for shared behavior across A2A data models.

    Provides a common configuration (e.g., alias-based population) and
    serves as the foundation for future extensions or shared utilities.

    This implementation provides backward compatibility for camelCase aliases
    by lazy-loading an alias map upon first use. Accessing or setting
    attributes via their camelCase alias will raise a DeprecationWarning.
    """

    model_config = ConfigDict(
        # SEE: https://docs.pydantic.dev/latest/api/config/#pydantic.config.ConfigDict.populate_by_name
        validate_by_name=True,
        validate_by_alias=True,
        serialize_by_alias=True,
        alias_generator=to_camel_custom,
    )



import asyncio
import contextlib
import logging

from types import TracebackType

from typing_extensions import Self

from a2a.server.events.event_queue import (
    DEFAULT_MAX_QUEUE_SIZE,
    AsyncQueue,
    Event,
    EventQueue,
    QueueShutDown,
    _create_async_queue,
)
from a2a.utils.telemetry import SpanKind, trace_class


logger = logging.getLogger(__name__)


@trace_class(kind=SpanKind.SERVER)
class EventQueueSource(EventQueue):
    """The Parent EventQueue.

    Acts as the single entry point for producers. Events pushed here are buffered
    in `_incoming_queue` and distributed to all child Sinks by a background dispatcher task.
    """

    def __init__(
        self,
        max_queue_size: int = DEFAULT_MAX_QUEUE_SIZE,
        create_default_sink: bool = True,
    ) -> None:
        """Initializes the EventQueueSource."""
        if max_queue_size <= 0:
            raise ValueError('max_queue_size must be greater than 0')

        self._incoming_queue: AsyncQueue[Event] = _create_async_queue(
            maxsize=max_queue_size
        )
        self._lock = asyncio.Lock()
        self._sinks: set[EventQueueSink] = set()
        self._is_closed = False

        # Internal sink for backward compatibility
        self._default_sink: EventQueueSink | None
        if create_default_sink:
            self._default_sink = EventQueueSink(
                parent=self, max_queue_size=max_queue_size
            )
            self._sinks.add(self._default_sink)
        else:
            self._default_sink = None

        self._dispatcher_task = asyncio.create_task(self._dispatch_loop())

        self._dispatcher_task_expected_to_cancel = False

        logger.debug('EventQueueSource initialized.')

    @property
    def queue(self) -> AsyncQueue[Event]:
        """Returns the underlying asyncio.Queue of the default sink."""
        if self._default_sink is None:
            raise ValueError('No default sink available.')
        return self._default_sink.queue

    async def _dispatch_loop(self) -> None:
        try:
            while True:
                event = await self._incoming_queue.get()

                async with self._lock:
                    active_sinks = list(self._sinks)

                if active_sinks:
                    results = await asyncio.gather(
                        *(
                            sink._put_internal(event)  # noqa: SLF001
                            for sink in active_sinks
                        ),
                        return_exceptions=True,
                    )
                    for result in results:
                        if isinstance(result, Exception):
                            logger.error(
                                'Error dispatching event to sink',
                                exc_info=result,
                            )

                self._incoming_queue.task_done()
        except asyncio.CancelledError:
            logger.debug(
                'EventQueueSource._dispatch_loop() for %s was cancelled',
                self,
            )
            if not self._dispatcher_task_expected_to_cancel:
                # This should only happen on forced shutdown (e.g. tests, server forced stop, etc).
                logger.info(
                    'EventQueueSource._dispatch_loop() for %s was cancelled without '
                    'calling EventQueue.close() first.',
                    self,
                )
                async with self._lock:
                    self._is_closed = True
                    sinks_to_close = list(self._sinks)

                self._incoming_queue.shutdown(immediate=True)
                await asyncio.gather(
                    *(sink.close(immediate=True) for sink in sinks_to_close)
                )
            raise
        except QueueShutDown:
            logger.debug('EventQueueSource._dispatch_loop() shutdown %s', self)
        except Exception:
            logger.exception(
                'EventQueueSource._dispatch_loop() failed %s', self
            )
            raise
        finally:
            logger.debug('EventQueueSource._dispatch_loop() Completed %s', self)

    async def _join_incoming_queue(self) -> None:
        """Helper to wait for join() while monitoring the dispatcher task."""
        if self._dispatcher_task.done():
            logger.warning(
                'Dispatcher task is not running. Cannot wait for event dispatch.'
            )
            return

        join_task = asyncio.create_task(self._incoming_queue.join())
        try:
            done, _pending = await asyncio.wait(
                [join_task, self._dispatcher_task],
                return_when=asyncio.FIRST_COMPLETED,
            )
        except asyncio.CancelledError:
            join_task.cancel()
            raise

        if join_task in done:
            return

        # Dispatcher task finished before join()
        join_task.cancel()
        with contextlib.suppress(asyncio.CancelledError):
            await join_task

        try:
            if self._dispatcher_task.exception():
                logger.error(
                    'Dispatcher task failed. Events may be lost.',
                    exc_info=self._dispatcher_task.exception(),
                )
            else:
                logger.warning(
                    'Dispatcher task finished unexpectedly. Events may be lost.'
                )
        except (asyncio.CancelledError, asyncio.InvalidStateError):
            logger.warning(
                'Dispatcher task was cancelled or finished. Events may be lost.'
            )

    async def tap(
        self, max_queue_size: int = DEFAULT_MAX_QUEUE_SIZE
    ) -> 'EventQueueSink':
        """Taps the event queue to create a new child queue that receives future events.

        Note: The tapped queue may receive some old events if the incoming event
        queue is lagging behind and hasn't dispatched them yet.
        """
        async with self._lock:
            if self._is_closed:
                raise QueueShutDown('Cannot tap a closed EventQueueSource.')
            sink = EventQueueSink(parent=self, max_queue_size=max_queue_size)
            self._sinks.add(sink)
            return sink

    async def remove_sink(self, sink: 'EventQueueSink') -> None:
        """Removes a sink from the source's internal list."""
        async with self._lock:
            self._sinks.remove(sink)

    async def enqueue_event(self, event: Event) -> None:
        """Enqueues an event to this queue and all its children."""
        logger.debug('Enqueuing event of type: %s', type(event))
        try:
            await self._incoming_queue.put(event)
        except QueueShutDown:
            logger.warning('Queue was closed during enqueuing. Event dropped.')
            return

    async def dequeue_event(self) -> Event:
        """Pulls an event from the default internal sink queue."""
        if self._default_sink is None:
            raise ValueError('No default sink available.')
        return await self._default_sink.dequeue_event()

    def task_done(self) -> None:
        """Signals that a work on dequeued event is complete via the default internal sink queue."""
        if self._default_sink is None:
            raise ValueError('No default sink available.')
        self._default_sink.task_done()

    async def close(self, immediate: bool = False) -> None:
        """Closes the queue and all its child sinks.

        It is safe to call it multiple times.
        If immediate is True, the queue will be closed without waiting for all events to be processed.
        If immediate is False, the queue will be closed after all events are processed (and confirmed with task_done() calls).

        WARNING: Closing the parent queue with immediate=False is a deadlock risk if there are unconsumed events
        in any of the child sinks and the consumer has crashed without draining its queue.
        It is highly recommended to wrap graceful shutdowns with a timeout, e.g.,
        `asyncio.wait_for(queue.close(immediate=False), timeout=...)`.
        """
        logger.debug('Closing EventQueueSource: immediate=%s', immediate)
        async with self._lock:
            # No more tap() allowed.
            self._is_closed = True
            # No more new events can be enqueued.
            self._incoming_queue.shutdown(immediate=immediate)
            sinks_to_close = list(self._sinks)

        if immediate:
            self._dispatcher_task_expected_to_cancel = True
            self._dispatcher_task.cancel()
            await asyncio.gather(
                *(sink.close(immediate=True) for sink in sinks_to_close)
            )
        else:
            # Wait for all already-enqueued events to be dispatched
            await self._join_incoming_queue()
            self._dispatcher_task_expected_to_cancel = True
            self._dispatcher_task.cancel()
            await asyncio.gather(
                *(sink.close(immediate=False) for sink in sinks_to_close)
            )

    def is_closed(self) -> bool:
        """[DEPRECATED] Checks if the queue is closed.

        NOTE: Relying on this for enqueue logic introduces race conditions.
        It is maintained primarily for backwards compatibility, workarounds for
        Python 3.10/3.12 async queues in consumers, and for the test suite.
        """
        return self._is_closed

    async def test_only_join_incoming_queue(self) -> None:
        """Wait for incoming queue to be fully processed."""
        await self._join_incoming_queue()

    async def __aenter__(self) -> Self:
        """Enters the async context manager, returning the queue itself.

        WARNING: See `__aexit__` for important deadlock risks associated with
        exiting this context manager if unconsumed events remain.
        """
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exits the async context manager, ensuring close() is called.

        WARNING: The context manager calls `close(immediate=False)` by default.
        If a consumer exits the `async with` block early (e.g., due to an exception
        or an explicit `break`) while unconsumed events remain in the queue,
        `__aexit__` will deadlock waiting for `task_done()` to be called on those events.
        """
        await self.close()


class EventQueueSink(EventQueue):
    """The Child EventQueue.

    Acts as a read-only consumer endpoint. Events are pushed here exclusively
    by the parent EventQueueSource's dispatcher task.
    """

    def __init__(
        self,
        parent: EventQueueSource,
        max_queue_size: int = DEFAULT_MAX_QUEUE_SIZE,
    ) -> None:
        """Initializes the EventQueueSink."""
        if max_queue_size <= 0:
            raise ValueError('max_queue_size must be greater than 0')

        self._parent = parent
        self._queue: AsyncQueue[Event] = _create_async_queue(
            maxsize=max_queue_size
        )
        self._is_closed = False
        self._lock = asyncio.Lock()

        logger.debug('EventQueueSink initialized.')

    @property
    def queue(self) -> AsyncQueue[Event]:
        """Returns the underlying asyncio.Queue of this sink."""
        return self._queue

    async def _put_internal(self, event: Event) -> None:
        with contextlib.suppress(QueueShutDown):
            await self._queue.put(event)

    async def enqueue_event(self, event: Event) -> None:
        """Sinks are read-only and cannot have events directly enqueued to them."""
        raise RuntimeError('Cannot enqueue to a sink-only queue')

    async def dequeue_event(self) -> Event:
        """Pulls an event from the sink queue."""
        logger.debug('Attempting to dequeue event (waiting).')
        event = await self._queue.get()
        logger.debug('Dequeued event: %s', event)
        return event

    def task_done(self) -> None:
        """Signals that a work on dequeued event is complete in this sink queue."""
        logger.debug('Marking task as done in EventQueueSink.')
        self._queue.task_done()

    async def tap(
        self, max_queue_size: int = DEFAULT_MAX_QUEUE_SIZE
    ) -> 'EventQueueSink':
        """Creates a child queue that receives future events.

        Note: The tapped queue may receive some old events if the incoming event
        queue is lagging behind and hasn't dispatched them yet.
        """
        # Delegate tap to the parent source so all sinks are flat under the source
        return await self._parent.tap(max_queue_size=max_queue_size)

    async def close(self, immediate: bool = False) -> None:
        """Closes the child sink queue.

        It is safe to call it multiple times.
        If immediate is True, the queue will be closed without waiting for all events to be processed.
        If immediate is False, the queue will be closed after all events are processed (and confirmed with task_done() calls).
        """
        logger.debug('Closing EventQueueSink.')
        async with self._lock:
            self._is_closed = True
            self._queue.shutdown(immediate=immediate)

        # Ignore KeyError (close have to be idempotent).
        with contextlib.suppress(KeyError):
            await self._parent.remove_sink(self)

        if not immediate:
            await self._queue.join()

    def is_closed(self) -> bool:
        """[DEPRECATED] Checks if the queue is closed.

        NOTE: Relying on this for enqueue logic introduces race conditions.
        It is maintained primarily for backwards compatibility, workarounds for
        Python 3.10/3.12 async queues in consumers, and for the test suite.
        """
        return self._is_closed

    async def __aenter__(self) -> Self:
        """Enters the async context manager, returning the queue itself.

        WARNING: See `__aexit__` for important deadlock risks associated with
        exiting this context manager if unconsumed events remain.
        """
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exits the async context manager, ensuring close() is called.

        WARNING: The context manager calls `close(immediate=False)` by default.
        If a consumer exits the `async with` block early (e.g., due to an exception
        or an explicit `break`) while unconsumed events remain in the queue,
        `__aexit__` will deadlock waiting for `task_done()` to be called on those events.
        """
        await self.close()



from __future__ import annotations

import asyncio  # noqa: TC003
import logging

from typing import TYPE_CHECKING, Any, cast

from a2a.server.agent_execution import (
    AgentExecutor,
    RequestContext,
    RequestContextBuilder,
    SimpleRequestContextBuilder,
)
from a2a.server.agent_execution.active_task import (
    INTERRUPTED_TASK_STATES,
    TERMINAL_TASK_STATES,
)
from a2a.server.agent_execution.active_task_registry import ActiveTaskRegistry
from a2a.server.request_handlers.request_handler import (
    RequestHandler,
    validate,
    validate_request_params,
)
from a2a.types.a2a_pb2 import (
    AgentCard,
    CancelTaskRequest,
    DeleteTaskPushNotificationConfigRequest,
    GetExtendedAgentCardRequest,
    GetTaskPushNotificationConfigRequest,
    GetTaskRequest,
    ListTaskPushNotificationConfigsRequest,
    ListTaskPushNotificationConfigsResponse,
    ListTasksRequest,
    ListTasksResponse,
    Message,
    SendMessageRequest,
    SubscribeToTaskRequest,
    Task,
    TaskPushNotificationConfig,
    TaskStatusUpdateEvent,
)
from a2a.utils.errors import (
    ExtendedAgentCardNotConfiguredError,
    InternalError,
    InvalidParamsError,
    PushNotificationNotSupportedError,
    TaskNotCancelableError,
    TaskNotFoundError,
)
from a2a.utils.task import (
    apply_history_length,
    validate_history_length,
    validate_page_size,
)
from a2a.utils.telemetry import SpanKind, trace_class


if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Awaitable, Callable

    from a2a.server.agent_execution.active_task import ActiveTask
    from a2a.server.context import ServerCallContext
    from a2a.server.events import Event
    from a2a.server.tasks import (
        PushNotificationConfigStore,
        PushNotificationSender,
        TaskStore,
    )


logger = logging.getLogger(__name__)


# TODO: cleanup context_id management


@trace_class(kind=SpanKind.SERVER)
class DefaultRequestHandlerV2(RequestHandler):
    """Default request handler for all incoming requests."""

    _background_tasks: set[asyncio.Task]

    def __init__(  # noqa: PLR0913
        self,
        agent_executor: AgentExecutor,
        task_store: TaskStore,
        agent_card: AgentCard,
        queue_manager: Any
        | None = None,  # Kept for backward compat in signature
        push_config_store: PushNotificationConfigStore | None = None,
        push_sender: PushNotificationSender | None = None,
        request_context_builder: RequestContextBuilder | None = None,
        extended_agent_card: AgentCard | None = None,
        extended_card_modifier: Callable[
            [AgentCard, ServerCallContext], Awaitable[AgentCard]
        ]
        | None = None,
    ) -> None:
        self.agent_executor = agent_executor
        self.task_store = task_store
        self._agent_card = agent_card
        self._push_config_store = push_config_store
        self._push_sender = push_sender
        self.extended_agent_card = extended_agent_card
        self.extended_card_modifier = extended_card_modifier
        self._request_context_builder = (
            request_context_builder
            or SimpleRequestContextBuilder(
                should_populate_referred_tasks=False, task_store=self.task_store
            )
        )
        self._active_task_registry = ActiveTaskRegistry(
            agent_executor=self.agent_executor,
            task_store=self.task_store,
            push_sender=self._push_sender,
        )
        self._background_tasks = set()

    @validate_request_params
    async def on_get_task(  # noqa: D102
        self,
        params: GetTaskRequest,
        context: ServerCallContext,
    ) -> Task | None:
        validate_history_length(params)

        task_id = params.id
        task: Task | None = await self.task_store.get(task_id, context)
        if not task:
            raise TaskNotFoundError

        return apply_history_length(task, params)

    @validate_request_params
    async def on_list_tasks(  # noqa: D102
        self,
        params: ListTasksRequest,
        context: ServerCallContext,
    ) -> ListTasksResponse:
        validate_history_length(params)
        if params.HasField('page_size'):
            validate_page_size(params.page_size)

        page = await self.task_store.list(params, context)
        for task in page.tasks:
            if not params.include_artifacts:
                task.ClearField('artifacts')

            updated_task = apply_history_length(task, params)
            if updated_task is not task:
                task.CopyFrom(updated_task)

        return page

    @validate_request_params
    async def on_cancel_task(  # noqa: D102
        self,
        params: CancelTaskRequest,
        context: ServerCallContext,
    ) -> Task | None:
        task_id = params.id

        try:
            active_task = await self._active_task_registry.get_or_create(
                task_id, call_context=context, create_task_if_missing=False
            )
            result = await active_task.cancel(context)
        except InvalidParamsError as e:
            raise TaskNotCancelableError from e

        if isinstance(result, Message):
            raise InternalError(
                message='Cancellation returned a message instead of a task.'
            )

        return result

    def _validate_task_id_match(self, task_id: str, event_task_id: str) -> None:
        if task_id != event_task_id:
            logger.error(
                'Agent generated task_id=%s does not match the RequestContext task_id=%s.',
                event_task_id,
                task_id,
            )
            raise InternalError(message='Task ID mismatch in agent response')

    async def _setup_active_task(
        self,
        params: SendMessageRequest,
        call_context: ServerCallContext,
    ) -> tuple[ActiveTask, RequestContext]:
        validate_history_length(params.configuration)

        original_task_id = params.message.task_id or None
        original_context_id = params.message.context_id or None

        if original_task_id:
            task = await self.task_store.get(original_task_id, call_context)
            if not task:
                raise TaskNotFoundError(f'Task {original_task_id} not found')

        # Build context to resolve or generate missing IDs
        request_context = await self._request_context_builder.build(
            params=params,
            task_id=original_task_id,
            context_id=original_context_id,
            # We will get the task when we have to process the request to avoid concurrent read/write issues.
            task=None,
            context=call_context,
        )

        task_id = cast('str', request_context.task_id)
        context_id = cast('str', request_context.context_id)

        if (
            self._push_config_store
            and params.configuration
            and params.configuration.task_push_notification_config
        ):
            await self._push_config_store.set_info(
                task_id,
                params.configuration.task_push_notification_config,
                call_context,
            )

        active_task = await self._active_task_registry.get_or_create(
            task_id,
            context_id=context_id,
            call_context=call_context,
            create_task_if_missing=True,
        )

        return active_task, request_context

    @validate_request_params
    async def on_message_send(  # noqa: D102
        self,
        params: SendMessageRequest,
        context: ServerCallContext,
    ) -> Message | Task:
        active_task, request_context = await self._setup_active_task(
            params, context
        )
        task_id = cast('str', request_context.task_id)

        result: Message | Task | None = None

        async for raw_event in active_task.subscribe(
            request=request_context,
            include_initial_task=False,
            replace_status_update_with_task=True,
        ):
            event = raw_event
            logger.debug(
                'Processing[%s] event [%s] %s',
                params.message.task_id,
                type(event).__name__,
                event,
            )
            if isinstance(event, TaskStatusUpdateEvent):
                self._validate_task_id_match(task_id, event.task_id)
                event = await active_task.get_task()
                logger.debug(
                    'Replaced TaskStatusUpdateEvent with Task: %s', event
                )

            if isinstance(event, Task) and (
                params.configuration.return_immediately
                or event.status.state
                in (TERMINAL_TASK_STATES | INTERRUPTED_TASK_STATES)
            ):
                self._validate_task_id_match(task_id, event.id)
                result = event
                # DO break here as it's "return_immediately".
                # AgentExecutor will continue to run in the background.
                break

            if isinstance(event, Message):
                result = event
                # Do NOT break here as Message is supposed to be the only
                # event in "Message-only" interaction.
                # ActiveTask consumer (see active_task.py) validates the event
                # stream and raises InvalidAgentResponseError if more events are
                # pushed after a Message.

        if result is None:
            logger.debug('Missing result for task %s', request_context.task_id)
            result = await active_task.get_task()

        if isinstance(result, Task):
            result = apply_history_length(result, params.configuration)

        logger.debug(
            'Returning result for task %s: %s',
            request_context.task_id,
            result,
        )
        return result

    @validate_request_params
    @validate(
        lambda self: self._agent_card.capabilities.streaming,
        'Streaming is not supported by the agent',
    )
    async def on_message_send_stream(  # noqa: D102
        self,
        params: SendMessageRequest,
        context: ServerCallContext,
    ) -> AsyncGenerator[Event, None]:
        active_task, request_context = await self._setup_active_task(
            params, context
        )

        task_id = cast('str', request_context.task_id)

        async for event in active_task.subscribe(
            request=request_context,
            include_initial_task=False,
        ):
            # Do NOT break here as we rely on AgentExecutor to yield control.
            # ActiveTask consumer (see active_task.py) validates the event
            # stream and raises InvalidAgentResponseError on misbehaving agents:
            #   - an event after a Message
            #   - Message after entering task mode
            #   - an event after a terminal state
            if isinstance(event, Task):
                self._validate_task_id_match(task_id, event.id)
                yield apply_history_length(event, params.configuration)
            else:
                yield event

    @validate_request_params
    @validate(
        lambda self: self._agent_card.capabilities.push_notifications,
        error_message='Push notifications are not supported by the agent',
        error_type=PushNotificationNotSupportedError,
    )
    async def on_create_task_push_notification_config(  # noqa: D102
        self,
        params: TaskPushNotificationConfig,
        context: ServerCallContext,
    ) -> TaskPushNotificationConfig:
        if not self._push_config_store:
            raise PushNotificationNotSupportedError

        task_id = params.task_id
        task: Task | None = await self.task_store.get(task_id, context)
        if not task:
            raise TaskNotFoundError

        await self._push_config_store.set_info(
            task_id,
            params,
            context,
        )

        return params

    @validate_request_params
    @validate(
        lambda self: self._agent_card.capabilities.push_notifications,
        error_message='Push notifications are not supported by the agent',
        error_type=PushNotificationNotSupportedError,
    )
    async def on_get_task_push_notification_config(  # noqa: D102
        self,
        params: GetTaskPushNotificationConfigRequest,
        context: ServerCallContext,
    ) -> TaskPushNotificationConfig:
        if not self._push_config_store:
            raise PushNotificationNotSupportedError

        task_id = params.task_id
        config_id = params.id
        task: Task | None = await self.task_store.get(task_id, context)
        if not task:
            raise TaskNotFoundError

        push_notification_configs: list[TaskPushNotificationConfig] = (
            await self._push_config_store.get_info(task_id, context) or []
        )

        for config in push_notification_configs:
            if config.id == config_id:
                return config

        raise TaskNotFoundError

    @validate_request_params
    @validate(
        lambda self: self._agent_card.capabilities.streaming,
        'Streaming is not supported by the agent',
    )
    async def on_subscribe_to_task(  # noqa: D102
        self,
        params: SubscribeToTaskRequest,
        context: ServerCallContext,
    ) -> AsyncGenerator[Event, None]:
        task_id = params.id

        active_task = await self._active_task_registry.get_or_create(
            task_id,
            call_context=context,
            create_task_if_missing=False,
        )

        async for event in active_task.subscribe(include_initial_task=True):
            yield event

    @validate_request_params
    @validate(
        lambda self: self._agent_card.capabilities.push_notifications,
        error_message='Push notifications are not supported by the agent',
        error_type=PushNotificationNotSupportedError,
    )
    async def on_list_task_push_notification_configs(  # noqa: D102
        self,
        params: ListTaskPushNotificationConfigsRequest,
        context: ServerCallContext,
    ) -> ListTaskPushNotificationConfigsResponse:
        if not self._push_config_store:
            raise PushNotificationNotSupportedError

        task_id = params.task_id
        task: Task | None = await self.task_store.get(task_id, context)
        if not task:
            raise TaskNotFoundError

        push_notification_config_list = await self._push_config_store.get_info(
            task_id, context
        )

        return ListTaskPushNotificationConfigsResponse(
            configs=push_notification_config_list
        )

    @validate_request_params
    @validate(
        lambda self: self._agent_card.capabilities.push_notifications,
        error_message='Push notifications are not supported by the agent',
        error_type=PushNotificationNotSupportedError,
    )
    async def on_delete_task_push_notification_config(  # noqa: D102
        self,
        params: DeleteTaskPushNotificationConfigRequest,
        context: ServerCallContext,
    ) -> None:
        if not self._push_config_store:
            raise PushNotificationNotSupportedError

        task_id = params.task_id
        config_id = params.id
        task: Task | None = await self.task_store.get(task_id, context)
        if not task:
            raise TaskNotFoundError

        await self._push_config_store.delete_info(task_id, context, config_id)

    @validate_request_params
    @validate(
        lambda self: self._agent_card.capabilities.extended_agent_card,
        error_message='The agent does not support authenticated extended cards',
    )
    async def on_get_extended_agent_card(
        self,
        params: GetExtendedAgentCardRequest,
        context: ServerCallContext,
    ) -> AgentCard:
        """Default handler for 'GetExtendedAgentCard'.

        Requires `capabilities.extended_agent_card` to be true.
        """
        extended_card = self.extended_agent_card
        if not extended_card:
            raise ExtendedAgentCardNotConfiguredError

        if self.extended_card_modifier:
            extended_card = await self.extended_card_modifier(
                extended_card, context
            )

        return extended_card



"""Custom exceptions for the A2A client."""

from a2a.utils.errors import A2AError


class A2AClientError(A2AError):
    """Base exception for A2A Client errors."""


class AgentCardResolutionError(A2AClientError):
    """Exception raised when an agent card cannot be resolved."""

    def __init__(self, message: str, status_code: int | None = None) -> None:
        super().__init__(message)
        self.status_code = status_code


class A2AClientTimeoutError(A2AClientError):
    """Exception for timeout errors during a request."""



import json
import logging

from collections.abc import AsyncGenerator
from typing import Any, NoReturn
from uuid import uuid4

import httpx

from jsonrpc.jsonrpc2 import JSONRPC20Request, JSONRPC20Response

from a2a.client.client import ClientCallContext
from a2a.client.errors import A2AClientError
from a2a.client.transports.base import ClientTransport
from a2a.client.transports.http_helpers import (
    get_http_args,
    send_http_request,
    send_http_stream_request,
)
from a2a.compat.v0_3 import conversions
from a2a.compat.v0_3 import types as types_v03
from a2a.compat.v0_3.extension_headers import add_legacy_extension_header
from a2a.types.a2a_pb2 import (
    AgentCard,
    CancelTaskRequest,
    DeleteTaskPushNotificationConfigRequest,
    GetExtendedAgentCardRequest,
    GetTaskPushNotificationConfigRequest,
    GetTaskRequest,
    ListTaskPushNotificationConfigsRequest,
    ListTaskPushNotificationConfigsResponse,
    ListTasksRequest,
    ListTasksResponse,
    SendMessageRequest,
    SendMessageResponse,
    StreamResponse,
    SubscribeToTaskRequest,
    Task,
    TaskPushNotificationConfig,
)
from a2a.utils.constants import PROTOCOL_VERSION_0_3, VERSION_HEADER
from a2a.utils.errors import JSON_RPC_ERROR_CODE_MAP
from a2a.utils.telemetry import SpanKind, trace_class


logger = logging.getLogger(__name__)

_JSON_RPC_ERROR_CODE_TO_A2A_ERROR = {
    code: error_type for error_type, code in JSON_RPC_ERROR_CODE_MAP.items()
}


@trace_class(kind=SpanKind.CLIENT)
class CompatJsonRpcTransport(ClientTransport):
    """A backward compatible JSON-RPC transport for A2A v0.3."""

    def __init__(
        self,
        httpx_client: httpx.AsyncClient,
        agent_card: AgentCard | None,
        url: str,
    ):
        """Initializes the CompatJsonRpcTransport."""
        self.url = url
        self.httpx_client = httpx_client
        self.agent_card = agent_card

    async def send_message(
        self,
        request: SendMessageRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> SendMessageResponse:
        """Sends a non-streaming message request to the agent."""
        req_v03 = conversions.to_compat_send_message_request(
            request, request_id=0
        )

        rpc_request = JSONRPC20Request(
            method='message/send',
            params=req_v03.params.model_dump(
                by_alias=True, exclude_none=True, mode='json'
            ),
            _id=str(uuid4()),
        )
        response_data = await self._send_request(
            dict(rpc_request.data), context
        )
        json_rpc_response = JSONRPC20Response(**response_data)
        if json_rpc_response.error:
            raise self._create_jsonrpc_error(json_rpc_response.error)

        result_dict = json_rpc_response.result
        if not isinstance(result_dict, dict):
            return SendMessageResponse()

        kind = result_dict.get('kind')

        # Fallback for old servers that might omit kind
        if not kind:
            if 'messageId' in result_dict:
                kind = 'message'
            elif 'id' in result_dict:
                kind = 'task'

        if kind == 'task':
            return SendMessageResponse(
                task=conversions.to_core_task(
                    types_v03.Task.model_validate(result_dict)
                )
            )
        if kind == 'message':
            return SendMessageResponse(
                message=conversions.to_core_message(
                    types_v03.Message.model_validate(result_dict)
                )
            )

        return SendMessageResponse()

    async def send_message_streaming(
        self,
        request: SendMessageRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> AsyncGenerator[StreamResponse]:
        """Sends a streaming message request to the agent and yields responses as they arrive."""
        req_v03 = conversions.to_compat_send_message_request(
            request, request_id=0
        )

        rpc_request = JSONRPC20Request(
            method='message/stream',
            params=req_v03.params.model_dump(
                by_alias=True, exclude_none=True, mode='json'
            ),
            _id=str(uuid4()),
        )
        async for event in self._send_stream_request(
            dict(rpc_request.data),
            context,
        ):
            yield event

    async def get_task(
        self,
        request: GetTaskRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> Task:
        """Retrieves the current state and history of a specific task."""
        req_v03 = conversions.to_compat_get_task_request(request, request_id=0)

        rpc_request = JSONRPC20Request(
            method='tasks/get',
            params=req_v03.params.model_dump(
                by_alias=True, exclude_none=True, mode='json'
            ),
            _id=str(uuid4()),
        )
        response_data = await self._send_request(
            dict(rpc_request.data), context
        )
        json_rpc_response = JSONRPC20Response(**response_data)
        if json_rpc_response.error:
            raise self._create_jsonrpc_error(json_rpc_response.error)
        return conversions.to_core_task(
            types_v03.Task.model_validate(json_rpc_response.result)
        )

    async def list_tasks(
        self,
        request: ListTasksRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> ListTasksResponse:
        """Retrieves tasks for an agent."""
        raise NotImplementedError(
            'ListTasks is not supported in A2A v0.3 JSONRPC.'
        )

    async def cancel_task(
        self,
        request: CancelTaskRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> Task:
        """Requests the agent to cancel a specific task."""
        req_v03 = conversions.to_compat_cancel_task_request(
            request, request_id=0
        )

        rpc_request = JSONRPC20Request(
            method='tasks/cancel',
            params=req_v03.params.model_dump(
                by_alias=True, exclude_none=True, mode='json'
            ),
            _id=str(uuid4()),
        )
        response_data = await self._send_request(
            dict(rpc_request.data), context
        )
        json_rpc_response = JSONRPC20Response(**response_data)
        if json_rpc_response.error:
            raise self._create_jsonrpc_error(json_rpc_response.error)

        return conversions.to_core_task(
            types_v03.Task.model_validate(json_rpc_response.result)
        )

    async def create_task_push_notification_config(
        self,
        request: TaskPushNotificationConfig,
        *,
        context: ClientCallContext | None = None,
    ) -> TaskPushNotificationConfig:
        """Sets or updates the push notification configuration for a specific task."""
        req_v03 = (
            conversions.to_compat_create_task_push_notification_config_request(
                request, request_id=0
            )
        )
        rpc_request = JSONRPC20Request(
            method='tasks/pushNotificationConfig/set',
            params=req_v03.params.model_dump(
                by_alias=True, exclude_none=True, mode='json'
            ),
            _id=str(uuid4()),
        )
        response_data = await self._send_request(
            dict(rpc_request.data), context
        )
        json_rpc_response = JSONRPC20Response(**response_data)
        if json_rpc_response.error:
            raise self._create_jsonrpc_error(json_rpc_response.error)

        return conversions.to_core_task_push_notification_config(
            types_v03.TaskPushNotificationConfig.model_validate(
                json_rpc_response.result
            )
        )

    async def get_task_push_notification_config(
        self,
        request: GetTaskPushNotificationConfigRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> TaskPushNotificationConfig:
        """Retrieves the push notification configuration for a specific task."""
        req_v03 = (
            conversions.to_compat_get_task_push_notification_config_request(
                request, request_id=0
            )
        )
        rpc_request = JSONRPC20Request(
            method='tasks/pushNotificationConfig/get',
            params=req_v03.params.model_dump(
                by_alias=True, exclude_none=True, mode='json'
            ),
            _id=str(uuid4()),
        )
        response_data = await self._send_request(
            dict(rpc_request.data), context
        )
        json_rpc_response = JSONRPC20Response(**response_data)
        if json_rpc_response.error:
            raise self._create_jsonrpc_error(json_rpc_response.error)

        return conversions.to_core_task_push_notification_config(
            types_v03.TaskPushNotificationConfig.model_validate(
                json_rpc_response.result
            )
        )

    async def list_task_push_notification_configs(
        self,
        request: ListTaskPushNotificationConfigsRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> ListTaskPushNotificationConfigsResponse:
        """Lists push notification configurations for a specific task."""
        req_v03 = (
            conversions.to_compat_list_task_push_notification_config_request(
                request, request_id=0
            )
        )
        rpc_request = JSONRPC20Request(
            method='tasks/pushNotificationConfig/list',
            params=req_v03.params.model_dump(
                by_alias=True, exclude_none=True, mode='json'
            ),
            _id=str(uuid4()),
        )
        response_data = await self._send_request(
            dict(rpc_request.data), context
        )
        json_rpc_response = JSONRPC20Response(**response_data)
        if json_rpc_response.error:
            raise self._create_jsonrpc_error(json_rpc_response.error)

        configs_data = json_rpc_response.result
        if not isinstance(configs_data, list):
            return ListTaskPushNotificationConfigsResponse()

        response = ListTaskPushNotificationConfigsResponse()
        for config_data in configs_data:
            response.configs.append(
                conversions.to_core_task_push_notification_config(
                    types_v03.TaskPushNotificationConfig.model_validate(
                        config_data
                    )
                )
            )
        return response

    async def delete_task_push_notification_config(
        self,
        request: DeleteTaskPushNotificationConfigRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> None:
        """Deletes the push notification configuration for a specific task."""
        req_v03 = (
            conversions.to_compat_delete_task_push_notification_config_request(
                request, request_id=0
            )
        )
        rpc_request = JSONRPC20Request(
            method='tasks/pushNotificationConfig/delete',
            params=req_v03.params.model_dump(
                by_alias=True, exclude_none=True, mode='json'
            ),
            _id=str(uuid4()),
        )
        response_data = await self._send_request(
            dict(rpc_request.data), context
        )
        if 'result' not in response_data and 'error' not in response_data:
            response_data['result'] = None

        json_rpc_response = JSONRPC20Response(**response_data)
        if json_rpc_response.error:
            raise self._create_jsonrpc_error(json_rpc_response.error)

    async def subscribe(
        self,
        request: SubscribeToTaskRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> AsyncGenerator[StreamResponse]:
        """Reconnects to get task updates."""
        req_v03 = conversions.to_compat_subscribe_to_task_request(
            request, request_id=0
        )
        rpc_request = JSONRPC20Request(
            method='tasks/resubscribe',
            params=req_v03.params.model_dump(
                by_alias=True, exclude_none=True, mode='json'
            ),
            _id=str(uuid4()),
        )
        async for event in self._send_stream_request(
            dict(rpc_request.data),
            context,
        ):
            yield event

    async def get_extended_agent_card(
        self,
        request: GetExtendedAgentCardRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> AgentCard:
        """Retrieves the Extended AgentCard."""
        card = self.agent_card
        if card and not card.capabilities.extended_agent_card:
            return card

        rpc_request = JSONRPC20Request(
            method='agent/getAuthenticatedExtendedCard',
            params={},
            _id=str(uuid4()),
        )
        response_data = await self._send_request(
            dict(rpc_request.data), context
        )
        json_rpc_response = JSONRPC20Response(**response_data)
        if json_rpc_response.error:
            raise self._create_jsonrpc_error(json_rpc_response.error)

        card = conversions.to_core_agent_card(
            types_v03.AgentCard.model_validate(json_rpc_response.result)
        )
        self.agent_card = card
        return card

    async def close(self) -> None:
        """Closes the httpx client."""
        await self.httpx_client.aclose()

    def _create_jsonrpc_error(
        self, error_dict: dict[str, Any]
    ) -> A2AClientError:
        """Raises a specific error based on jsonrpc error code."""
        code = error_dict.get('code')
        message = error_dict.get('message', 'Unknown Error')

        if isinstance(code, int):
            error_class = _JSON_RPC_ERROR_CODE_TO_A2A_ERROR.get(code)
            if error_class:
                return error_class(message)  # type: ignore[return-value]

        return A2AClientError(message)

    def _handle_http_error(self, e: httpx.HTTPStatusError) -> NoReturn:
        """Handles HTTP errors for standard requests."""
        raise A2AClientError(f'HTTP Error: {e.response.status_code}') from e

    async def _send_stream_request(
        self,
        json_data: dict[str, Any],
        context: ClientCallContext | None = None,
    ) -> AsyncGenerator[StreamResponse]:
        """Sends an HTTP stream request."""
        http_kwargs = get_http_args(context)
        http_kwargs.setdefault('headers', {})
        http_kwargs['headers'][VERSION_HEADER.lower()] = PROTOCOL_VERSION_0_3
        add_legacy_extension_header(http_kwargs['headers'])

        async for sse_data in send_http_stream_request(
            self.httpx_client,
            'POST',
            self.url,
            self._handle_http_error,
            json=json_data,
            **http_kwargs,
        ):
            data = json.loads(sse_data)
            if 'error' in data:
                raise self._create_jsonrpc_error(data['error'])

            result_dict = data.get('result', {})
            if not isinstance(result_dict, dict):
                continue

            kind = result_dict.get('kind')

            if not kind:
                if 'taskId' in result_dict and 'final' in result_dict:
                    kind = 'status-update'
                elif 'messageId' in result_dict:
                    kind = 'message'
                elif 'id' in result_dict:
                    kind = 'task'

            result: (
                types_v03.Task
                | types_v03.Message
                | types_v03.TaskStatusUpdateEvent
                | types_v03.TaskArtifactUpdateEvent
            )
            if kind == 'task':
                result = types_v03.Task.model_validate(result_dict)
            elif kind == 'message':
                result = types_v03.Message.model_validate(result_dict)
            elif kind == 'status-update':
                result = types_v03.TaskStatusUpdateEvent.model_validate(
                    result_dict
                )
            elif kind == 'artifact-update':
                result = types_v03.TaskArtifactUpdateEvent.model_validate(
                    result_dict
                )
            else:
                continue

            yield conversions.to_core_stream_response(
                types_v03.SendStreamingMessageSuccessResponse(result=result)
            )

    async def _send_request(
        self,
        json_data: dict[str, Any],
        context: ClientCallContext | None = None,
    ) -> dict[str, Any]:
        """Sends an HTTP request."""
        http_kwargs = get_http_args(context)
        http_kwargs.setdefault('headers', {})
        http_kwargs['headers'][VERSION_HEADER.lower()] = PROTOCOL_VERSION_0_3
        add_legacy_extension_header(http_kwargs['headers'])

        request = self.httpx_client.build_request(
            'POST',
            self.url,
            json=json_data,
            **http_kwargs,
        )
        return await send_http_request(
            self.httpx_client, request, self._handle_http_error
        )



from a2a.types.a2a_pb2 import AgentCard, AgentExtension


HTTP_EXTENSION_HEADER = 'A2A-Extensions'


def get_requested_extensions(values: list[str]) -> set[str]:
    """Get the set of requested extensions from an input list.

    This handles the list containing potentially comma-separated values, as
    occurs when using a list in an HTTP header.
    """
    return {
        stripped
        for v in values
        for ext in v.split(',')
        if (stripped := ext.strip())
    }


def find_extension_by_uri(card: AgentCard, uri: str) -> AgentExtension | None:
    """Find an AgentExtension in an AgentCard given a uri."""
    for ext in card.capabilities.extensions or []:
        if ext.uri == uri:
            return ext

    return None



from typing import Any, Literal

from pydantic import BaseModel


class JSONRPCBaseModel(BaseModel):
    """Base model for JSON-RPC objects."""

    model_config = {
        'extra': 'allow',
        'populate_by_name': True,
        'arbitrary_types_allowed': True,
    }


class JSONRPCError(JSONRPCBaseModel):
    """Base model for JSON-RPC error objects."""

    code: int
    message: str
    data: Any | None = None


class JSONParseError(JSONRPCError):
    """Error raised when invalid JSON was received by the server."""

    code: Literal[-32700] = -32700  # pyright: ignore [reportIncompatibleVariableOverride]
    message: str = 'Parse error'


class InvalidRequestError(JSONRPCError):
    """Error raised when the JSON sent is not a valid Request object."""

    code: Literal[-32600] = -32600  # pyright: ignore [reportIncompatibleVariableOverride]
    message: str = 'Invalid Request'


class MethodNotFoundError(JSONRPCError):
    """Error raised when the method does not exist / is not available."""

    code: Literal[-32601] = -32601  # pyright: ignore [reportIncompatibleVariableOverride]
    message: str = 'Method not found'


class InvalidParamsError(JSONRPCError):
    """Error raised when invalid method parameter(s)."""

    code: Literal[-32602] = -32602  # pyright: ignore [reportIncompatibleVariableOverride]
    message: str = 'Invalid params'


class InternalError(JSONRPCError):
    """Error raised when internal JSON-RPC error."""

    code: Literal[-32603] = -32603  # pyright: ignore [reportIncompatibleVariableOverride]
    message: str = 'Internal error'



import logging
import typing

from collections.abc import AsyncIterable

from a2a.compat.v0_3 import conversions
from a2a.compat.v0_3 import types as types_v03
from a2a.server.context import ServerCallContext
from a2a.server.request_handlers.request_handler import RequestHandler
from a2a.types.a2a_pb2 import Task
from a2a.utils import proto_utils as core_proto_utils
from a2a.utils.errors import TaskNotFoundError


logger = logging.getLogger(__name__)


class RequestHandler03:
    """A protocol-agnostic v0.3 RequestHandler that delegates to the v1.0 RequestHandler."""

    def __init__(self, request_handler: RequestHandler):
        self.request_handler = request_handler

    async def on_message_send(
        self,
        request: types_v03.SendMessageRequest,
        context: ServerCallContext,
    ) -> types_v03.Task | types_v03.Message:
        """Sends a message using v0.3 protocol types."""
        v10_req = conversions.to_core_send_message_request(request)
        task_or_message = await self.request_handler.on_message_send(
            v10_req, context
        )
        if isinstance(task_or_message, Task):
            return conversions.to_compat_task(task_or_message)
        return conversions.to_compat_message(task_or_message)

    async def on_message_send_stream(
        self,
        request: types_v03.SendMessageRequest,
        context: ServerCallContext,
    ) -> AsyncIterable[types_v03.SendStreamingMessageSuccessResponse]:
        """Sends a message stream using v0.3 protocol types."""
        v10_req = conversions.to_core_send_message_request(request)
        async for event in self.request_handler.on_message_send_stream(
            v10_req, context
        ):
            v10_stream_resp = core_proto_utils.to_stream_response(event)
            yield conversions.to_compat_stream_response(
                v10_stream_resp, request.id
            )

    async def on_cancel_task(
        self,
        request: types_v03.CancelTaskRequest,
        context: ServerCallContext,
    ) -> types_v03.Task:
        """Cancels a task using v0.3 protocol types."""
        v10_req = conversions.to_core_cancel_task_request(request)
        v10_task = await self.request_handler.on_cancel_task(v10_req, context)
        if v10_task:
            return conversions.to_compat_task(v10_task)
        raise TaskNotFoundError

    async def on_subscribe_to_task(
        self,
        request: types_v03.TaskResubscriptionRequest,
        context: ServerCallContext,
    ) -> AsyncIterable[types_v03.SendStreamingMessageSuccessResponse]:
        """Subscribes to a task using v0.3 protocol types."""
        v10_req = conversions.to_core_subscribe_to_task_request(request)
        async for event in self.request_handler.on_subscribe_to_task(
            v10_req, context
        ):
            v10_stream_resp = core_proto_utils.to_stream_response(event)
            yield conversions.to_compat_stream_response(
                v10_stream_resp, request.id
            )

    async def on_get_task_push_notification_config(
        self,
        request: types_v03.GetTaskPushNotificationConfigRequest,
        context: ServerCallContext,
    ) -> types_v03.TaskPushNotificationConfig:
        """Gets a push notification config using v0.3 protocol types."""
        v10_req = conversions.to_core_get_task_push_notification_config_request(
            request
        )
        v10_config = (
            await self.request_handler.on_get_task_push_notification_config(
                v10_req, context
            )
        )
        return conversions.to_compat_task_push_notification_config(v10_config)

    async def on_create_task_push_notification_config(
        self,
        request: types_v03.SetTaskPushNotificationConfigRequest,
        context: ServerCallContext,
    ) -> types_v03.TaskPushNotificationConfig:
        """Creates a push notification config using v0.3 protocol types."""
        v10_req = (
            conversions.to_core_create_task_push_notification_config_request(
                request
            )
        )
        v10_config = (
            await self.request_handler.on_create_task_push_notification_config(
                v10_req, context
            )
        )
        return conversions.to_compat_task_push_notification_config(v10_config)

    async def on_get_task(
        self,
        request: types_v03.GetTaskRequest,
        context: ServerCallContext,
    ) -> types_v03.Task:
        """Gets a task using v0.3 protocol types."""
        v10_req = conversions.to_core_get_task_request(request)
        v10_task = await self.request_handler.on_get_task(v10_req, context)
        if v10_task:
            return conversions.to_compat_task(v10_task)
        raise TaskNotFoundError

    async def on_list_task_push_notification_configs(
        self,
        request: types_v03.ListTaskPushNotificationConfigRequest,
        context: ServerCallContext,
    ) -> list[types_v03.TaskPushNotificationConfig]:
        """Lists push notification configs using v0.3 protocol types."""
        v10_req = (
            conversions.to_core_list_task_push_notification_config_request(
                request
            )
        )
        v10_resp = (
            await self.request_handler.on_list_task_push_notification_configs(
                v10_req, context
            )
        )
        v03_resp = (
            conversions.to_compat_list_task_push_notification_config_response(
                v10_resp, request.id
            )
        )
        if isinstance(
            v03_resp.root,
            types_v03.ListTaskPushNotificationConfigSuccessResponse,
        ):
            return typing.cast(
                'list[types_v03.TaskPushNotificationConfig]',
                v03_resp.root.result,
            )
        return []

    async def on_delete_task_push_notification_config(
        self,
        request: types_v03.DeleteTaskPushNotificationConfigRequest,
        context: ServerCallContext,
    ) -> None:
        """Deletes a push notification config using v0.3 protocol types."""
        v10_req = (
            conversions.to_core_delete_task_push_notification_config_request(
                request
            )
        )
        await self.request_handler.on_delete_task_push_notification_config(
            v10_req, context
        )

    async def on_get_extended_agent_card(
        self,
        request: types_v03.GetAuthenticatedExtendedCardRequest,
        context: ServerCallContext,
    ) -> types_v03.AgentCard:
        """Gets the authenticated extended agent card using v0.3 protocol types."""
        v10_req = conversions.to_core_get_extended_agent_card_request(request)
        v10_card = await self.request_handler.on_get_extended_agent_card(
            v10_req, context
        )
        return conversions.to_compat_agent_card(v10_card)



"""Client-side components for interacting with an A2A agent."""

from a2a.client.auth import (
    AuthInterceptor,
    CredentialService,
    InMemoryContextCredentialStore,
)
from a2a.client.base_client import BaseClient
from a2a.client.card_resolver import A2ACardResolver
from a2a.client.client import (
    Client,
    ClientCallContext,
    ClientConfig,
)
from a2a.client.client_factory import (
    ClientFactory,
    create_client,
    minimal_agent_card,
)
from a2a.client.errors import (
    A2AClientError,
    A2AClientTimeoutError,
    AgentCardResolutionError,
)
from a2a.client.interceptors import ClientCallInterceptor


__all__ = [
    'A2ACardResolver',
    'A2AClientError',
    'A2AClientTimeoutError',
    'AgentCardResolutionError',
    'AuthInterceptor',
    'BaseClient',
    'Client',
    'ClientCallContext',
    'ClientCallInterceptor',
    'ClientConfig',
    'ClientFactory',
    'CredentialService',
    'InMemoryContextCredentialStore',
    'create_client',
    'minimal_agent_card',
]



"""Custom exceptions and error types for A2A server-side errors.

This module contains A2A-specific error codes,
as well as server exception classes.
"""

from typing import NamedTuple


class RestErrorMap(NamedTuple):
    """Named tuple mapping HTTP status, gRPC status, and reason strings."""

    http_code: int
    grpc_status: str
    reason: str


class A2AError(Exception):
    """Base exception for A2A errors."""

    message: str = 'A2A Error'
    data: dict | None = None

    def __init__(self, message: str | None = None, data: dict | None = None):
        if message:
            self.message = message
        self.data = data
        super().__init__(self.message)


class TaskNotFoundError(A2AError):
    """Exception raised when a task is not found."""

    message = 'Task not found'


class TaskNotCancelableError(A2AError):
    """Exception raised when a task cannot be canceled."""

    message = 'Task cannot be canceled'


class PushNotificationNotSupportedError(A2AError):
    """Exception raised when push notifications are not supported."""

    message = 'Push Notification is not supported'


class UnsupportedOperationError(A2AError):
    """Exception raised when an operation is not supported."""

    message = 'This operation is not supported'


class ContentTypeNotSupportedError(A2AError):
    """Exception raised when the content type is incompatible."""

    message = 'Incompatible content types'


class InternalError(A2AError):
    """Exception raised for internal server errors."""

    message = 'Internal error'


class InvalidAgentResponseError(A2AError):
    """Exception raised when the agent response is invalid."""

    message = 'Invalid agent response'


class ExtendedAgentCardNotConfiguredError(A2AError):
    """Exception raised when the authenticated extended card is not configured."""

    message = 'Authenticated Extended Card is not configured'


class InvalidParamsError(A2AError):
    """Exception raised when parameters are invalid."""

    message = 'Invalid params'


class InvalidRequestError(A2AError):
    """Exception raised when the request is invalid."""

    message = 'Invalid Request'


class MethodNotFoundError(A2AError):
    """Exception raised when a method is not found."""

    message = 'Method not found'


class ExtensionSupportRequiredError(A2AError):
    """Exception raised when extension support is required but not present."""

    message = 'Extension support required'


class VersionNotSupportedError(A2AError):
    """Exception raised when the requested version is not supported."""

    message = 'Version not supported'


# For backward compatibility if needed, or just aliases for clean refactor
# We remove the Pydantic models here.

__all__ = [
    'A2A_ERROR_REASONS',
    'A2A_REASON_TO_ERROR',
    'A2A_REST_ERROR_MAPPING',
    'JSON_RPC_ERROR_CODE_MAP',
    'ExtensionSupportRequiredError',
    'InternalError',
    'InvalidAgentResponseError',
    'InvalidParamsError',
    'InvalidRequestError',
    'MethodNotFoundError',
    'PushNotificationNotSupportedError',
    'RestErrorMap',
    'TaskNotCancelableError',
    'TaskNotFoundError',
    'UnsupportedOperationError',
    'VersionNotSupportedError',
]


JSON_RPC_ERROR_CODE_MAP: dict[type[A2AError], int] = {
    TaskNotFoundError: -32001,
    TaskNotCancelableError: -32002,
    PushNotificationNotSupportedError: -32003,
    UnsupportedOperationError: -32004,
    ContentTypeNotSupportedError: -32005,
    InvalidAgentResponseError: -32006,
    ExtendedAgentCardNotConfiguredError: -32007,
    ExtensionSupportRequiredError: -32008,
    VersionNotSupportedError: -32009,
    InvalidParamsError: -32602,
    InvalidRequestError: -32600,
    MethodNotFoundError: -32601,
    InternalError: -32603,
}


A2A_REST_ERROR_MAPPING: dict[type[A2AError], RestErrorMap] = {
    TaskNotFoundError: RestErrorMap(404, 'NOT_FOUND', 'TASK_NOT_FOUND'),
    TaskNotCancelableError: RestErrorMap(
        409, 'FAILED_PRECONDITION', 'TASK_NOT_CANCELABLE'
    ),
    PushNotificationNotSupportedError: RestErrorMap(
        400,
        'UNIMPLEMENTED',
        'PUSH_NOTIFICATION_NOT_SUPPORTED',
    ),
    UnsupportedOperationError: RestErrorMap(
        400, 'UNIMPLEMENTED', 'UNSUPPORTED_OPERATION'
    ),
    ContentTypeNotSupportedError: RestErrorMap(
        415,
        'INVALID_ARGUMENT',
        'CONTENT_TYPE_NOT_SUPPORTED',
    ),
    InvalidAgentResponseError: RestErrorMap(
        502, 'INTERNAL', 'INVALID_AGENT_RESPONSE'
    ),
    ExtendedAgentCardNotConfiguredError: RestErrorMap(
        400,
        'FAILED_PRECONDITION',
        'EXTENDED_AGENT_CARD_NOT_CONFIGURED',
    ),
    ExtensionSupportRequiredError: RestErrorMap(
        400,
        'FAILED_PRECONDITION',
        'EXTENSION_SUPPORT_REQUIRED',
    ),
    VersionNotSupportedError: RestErrorMap(
        400, 'UNIMPLEMENTED', 'VERSION_NOT_SUPPORTED'
    ),
    InvalidParamsError: RestErrorMap(400, 'INVALID_ARGUMENT', 'INVALID_PARAMS'),
    InvalidRequestError: RestErrorMap(
        400, 'INVALID_ARGUMENT', 'INVALID_REQUEST'
    ),
    MethodNotFoundError: RestErrorMap(404, 'NOT_FOUND', 'METHOD_NOT_FOUND'),
    InternalError: RestErrorMap(500, 'INTERNAL', 'INTERNAL_ERROR'),
}


A2A_ERROR_REASONS = {
    cls: mapping.reason for cls, mapping in A2A_REST_ERROR_MAPPING.items()
}

A2A_REASON_TO_ERROR = {
    mapping.reason: cls for cls, mapping in A2A_REST_ERROR_MAPPING.items()
}



import argparse
import asyncio
import contextlib
import logging

import grpc
import uvicorn

from fastapi import FastAPI

from a2a.compat.v0_3 import a2a_v0_3_pb2_grpc
from a2a.compat.v0_3.grpc_handler import CompatGrpcHandler
from a2a.server.agent_execution.agent_executor import AgentExecutor
from a2a.server.agent_execution.context import RequestContext
from a2a.server.events.event_queue import EventQueue
from a2a.server.request_handlers import DefaultRequestHandler, GrpcHandler
from a2a.server.routes import (
    create_agent_card_routes,
    create_jsonrpc_routes,
    create_rest_routes,
)
from a2a.server.tasks.inmemory_task_store import InMemoryTaskStore
from a2a.server.tasks.task_updater import TaskUpdater
from a2a.types import (
    AgentCapabilities,
    AgentCard,
    AgentInterface,
    AgentProvider,
    AgentSkill,
    Part,
    Task,
    TaskState,
    TaskStatus,
    a2a_pb2_grpc,
)


logger = logging.getLogger(__name__)


class SampleAgentExecutor(AgentExecutor):
    """Sample agent executor logic similar to the a2a-js sample."""

    def __init__(self) -> None:
        self.running_tasks: set[str] = set()

    async def cancel(
        self, context: RequestContext, event_queue: EventQueue
    ) -> None:
        """Cancels a task."""
        task_id = context.task_id
        if task_id in self.running_tasks:
            self.running_tasks.remove(task_id)

        updater = TaskUpdater(
            event_queue=event_queue,
            task_id=task_id or '',
            context_id=context.context_id or '',
        )
        await updater.cancel()

    async def execute(
        self, context: RequestContext, event_queue: EventQueue
    ) -> None:
        """Executes a task inline."""
        user_message = context.message
        task_id = context.task_id
        context_id = context.context_id

        if not user_message or not task_id or not context_id:
            return

        self.running_tasks.add(task_id)

        logger.info(
            '[SampleAgentExecutor] Processing message %s for task %s (context: %s)',
            user_message.message_id,
            task_id,
            context_id,
        )

        await event_queue.enqueue_event(
            Task(
                id=task_id,
                context_id=context_id,
                status=TaskStatus(state=TaskState.TASK_STATE_SUBMITTED),
                history=[user_message],
            )
        )

        updater = TaskUpdater(
            event_queue=event_queue,
            task_id=task_id,
            context_id=context_id,
        )

        working_message = updater.new_agent_message(
            parts=[Part(text='Processing your question...')]
        )
        await updater.start_work(message=working_message)

        query = context.get_user_input()

        agent_reply_text = self._parse_input(query)
        await asyncio.sleep(1)

        if task_id not in self.running_tasks:
            return

        await updater.add_artifact(
            parts=[Part(text=agent_reply_text)],
            name='response',
            last_chunk=True,
        )
        await updater.complete()

        logger.info(
            '[SampleAgentExecutor] Task %s finished with state: completed',
            task_id,
        )

    def _parse_input(self, query: str) -> str:
        if not query:
            return 'Hello! Please provide a message for me to respond to.'

        ql = query.lower()
        if 'hello' in ql or 'hi' in ql:
            return 'Hello World! Nice to meet you!'
        if 'how are you' in ql:
            return (
                "I'm doing great! Thanks for asking. How can I help you today?"
            )
        if 'goodbye' in ql or 'bye' in ql:
            return 'Goodbye! Have a wonderful day!'
        return f"Hello World! You said: '{query}'. Thanks for your message!"


async def serve(
    host: str = '127.0.0.1',
    port: int = 41241,
    grpc_port: int = 50051,
    compat_grpc_port: int = 50052,
) -> None:
    """Run the Sample Agent server with mounted JSON-RPC, HTTP+JSON and gRPC transports."""
    agent_card = AgentCard(
        name='Sample Agent',
        description='A sample agent to test the stream functionality.',
        provider=AgentProvider(
            organization='A2A Samples', url='https://example.com'
        ),
        version='1.0.0',
        capabilities=AgentCapabilities(
            streaming=True, push_notifications=False
        ),
        default_input_modes=['text'],
        default_output_modes=['text', 'task-status'],
        skills=[
            AgentSkill(
                id='sample_agent',
                name='Sample Agent',
                description='Say hi.',
                tags=['sample'],
                examples=['hi'],
                input_modes=['text'],
                output_modes=['text', 'task-status'],
            )
        ],
        supported_interfaces=[
            AgentInterface(
                protocol_binding='GRPC',
                protocol_version='1.0',
                url=f'{host}:{grpc_port}',
            ),
            AgentInterface(
                protocol_binding='GRPC',
                protocol_version='0.3',
                url=f'{host}:{compat_grpc_port}',
            ),
            AgentInterface(
                protocol_binding='JSONRPC',
                protocol_version='1.0',
                url=f'http://{host}:{port}/a2a/jsonrpc',
            ),
            AgentInterface(
                protocol_binding='JSONRPC',
                protocol_version='0.3',
                url=f'http://{host}:{port}/a2a/jsonrpc',
            ),
            AgentInterface(
                protocol_binding='HTTP+JSON',
                protocol_version='1.0',
                url=f'http://{host}:{port}/a2a/rest',
            ),
            AgentInterface(
                protocol_binding='HTTP+JSON',
                protocol_version='0.3',
                url=f'http://{host}:{port}/a2a/rest',
            ),
        ],
    )

    task_store = InMemoryTaskStore()
    request_handler = DefaultRequestHandler(
        agent_executor=SampleAgentExecutor(),
        task_store=task_store,
        agent_card=agent_card,
    )

    rest_routes = create_rest_routes(
        request_handler=request_handler,
        path_prefix='/a2a/rest',
        enable_v0_3_compat=True,
    )
    jsonrpc_routes = create_jsonrpc_routes(
        request_handler=request_handler,
        rpc_url='/a2a/jsonrpc',
        enable_v0_3_compat=True,
    )
    agent_card_routes = create_agent_card_routes(
        agent_card=agent_card,
    )
    app = FastAPI()
    app.routes.extend(jsonrpc_routes)
    app.routes.extend(agent_card_routes)
    app.routes.extend(rest_routes)

    grpc_server = grpc.aio.server()
    grpc_server.add_insecure_port(f'{host}:{grpc_port}')
    servicer = GrpcHandler(request_handler)
    a2a_pb2_grpc.add_A2AServiceServicer_to_server(servicer, grpc_server)

    compat_grpc_server = grpc.aio.server()
    compat_grpc_server.add_insecure_port(f'{host}:{compat_grpc_port}')
    compat_servicer = CompatGrpcHandler(request_handler)
    a2a_v0_3_pb2_grpc.add_A2AServiceServicer_to_server(
        compat_servicer, compat_grpc_server
    )

    config = uvicorn.Config(app, host=host, port=port)
    uvicorn_server = uvicorn.Server(config)

    logger.info('Starting Sample Agent servers:')
    logger.info(' - HTTP on http://%s:%s', host, port)
    logger.info(' - gRPC on %s:%s', host, grpc_port)
    logger.info(' - gRPC (v0.3 compat) on %s:%s', host, compat_grpc_port)
    logger.info(
        'Agent Card available at http://%s:%s/.well-known/agent-card.json',
        host,
        port,
    )

    await asyncio.gather(
        grpc_server.start(),
        compat_grpc_server.start(),
        uvicorn_server.serve(),
    )


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    parser = argparse.ArgumentParser(description='Sample A2A agent server')
    parser.add_argument('--host', default='127.0.0.1')
    parser.add_argument('--port', type=int, default=41241)
    parser.add_argument('--grpc-port', type=int, default=50051)
    parser.add_argument('--compat-grpc-port', type=int, default=50052)
    args = parser.parse_args()
    with contextlib.suppress(KeyboardInterrupt):
        asyncio.run(
            serve(
                host=args.host,
                port=args.port,
                grpc_port=args.grpc_port,
                compat_grpc_port=args.compat_grpc_port,
            )
        )



import argparse
import asyncio
import os
import signal
import uuid

from typing import Any

import grpc
import httpx

from a2a.client import A2ACardResolver, ClientConfig, create_client
from a2a.helpers import get_artifact_text, get_message_text
from a2a.helpers.agent_card import display_agent_card
from a2a.types import Message, Part, Role, SendMessageRequest, TaskState


async def _handle_stream(
    stream: Any, current_task_id: str | None
) -> str | None:
    async for event in stream:
        if event.HasField('message'):
            print('Message:', get_message_text(event.message, delimiter=' '))
            return None

        if not current_task_id:
            if event.HasField('task'):
                current_task_id = event.task.id
                print('--- Task Started ---')
                print(f'Task [state={TaskState.Name(event.task.status.state)}]')
            else:
                raise ValueError(f'Unexpected first event: {event}')

        if event.HasField('status_update'):
            state_name = TaskState.Name(event.status_update.status.state)
            message_text = (
                ': '
                + get_message_text(
                    event.status_update.status.message, delimiter=' '
                )
                if event.status_update.status.HasField('message')
                else ''
            )
            print(f'TaskStatusUpdate [state={state_name}]{message_text}')
            if state_name in (
                'TASK_STATE_COMPLETED',
                'TASK_STATE_FAILED',
                'TASK_STATE_CANCELED',
                'TASK_STATE_REJECTED',
            ):
                current_task_id = None
                print('--- Task Finished ---')
        elif event.HasField('artifact_update'):
            print(
                f'TaskArtifactUpdate [name={event.artifact_update.artifact.name}]:',
                get_artifact_text(
                    event.artifact_update.artifact, delimiter=' '
                ),
            )
    return current_task_id


async def main() -> None:
    """Run the A2A terminal client."""
    parser = argparse.ArgumentParser(description='A2A Terminal Client')
    parser.add_argument(
        '--url', default='http://127.0.0.1:41241', help='Agent base URL'
    )
    parser.add_argument(
        '--transport',
        default=None,
        help='Preferred transport (JSONRPC, HTTP+JSON, GRPC)',
    )
    args = parser.parse_args()

    config = ClientConfig(
        grpc_channel_factory=grpc.aio.insecure_channel,
    )
    if args.transport:
        config.supported_protocol_bindings = [args.transport]

    print(
        f'Connecting to {args.url} (preferred transport: {args.transport or "Any"})'
    )

    async with httpx.AsyncClient() as httpx_client:
        resolver = A2ACardResolver(httpx_client, args.url)
        card = await resolver.get_agent_card()
        print('\n✓ Agent Card Found:')
        display_agent_card(card)

    client = await create_client(card, client_config=config)

    actual_transport = getattr(client, '_transport', client)
    print(f'  Picked Transport: {actual_transport.__class__.__name__}')

    print('\nConnected! Send a message or type /quit to exit.')

    current_task_id = None
    current_context_id = str(uuid.uuid4())

    while True:
        try:
            loop = asyncio.get_running_loop()
            user_input = await loop.run_in_executor(None, input, 'You: ')
        except KeyboardInterrupt:
            break

        if user_input.lower() in ('/quit', '/exit'):
            break
        if not user_input.strip():
            continue

        message = Message(
            role=Role.ROLE_USER,
            message_id=str(uuid.uuid4()),
            parts=[Part(text=user_input)],
            task_id=current_task_id,
            context_id=current_context_id,
        )

        request = SendMessageRequest(message=message)

        try:
            stream = client.send_message(request)
            current_task_id = await _handle_stream(stream, current_task_id)
        except (httpx.RequestError, grpc.RpcError) as e:
            print(f'Error communicating with agent: {e}')

    await client.close()


if __name__ == '__main__':
    signal.signal(signal.SIGINT, lambda sig, frame: os._exit(0))
    asyncio.run(main())



"""A2A Routes."""

from a2a.server.routes.agent_card_routes import create_agent_card_routes
from a2a.server.routes.common import (
    DefaultServerCallContextBuilder,
    ServerCallContextBuilder,
)
from a2a.server.routes.jsonrpc_routes import create_jsonrpc_routes
from a2a.server.routes.rest_routes import create_rest_routes


__all__ = [
    'DefaultServerCallContextBuilder',
    'ServerCallContextBuilder',
    'create_agent_card_routes',
    'create_jsonrpc_routes',
    'create_rest_routes',
]



# ruff: noqa: PLC0415
import logging

from typing import TYPE_CHECKING

from google.protobuf.json_format import MessageToJson, Parse


try:
    from sqlalchemy import ColumnElement, Table, and_, delete, select
    from sqlalchemy.ext.asyncio import (
        AsyncEngine,
        AsyncSession,
        async_sessionmaker,
    )
    from sqlalchemy.orm import class_mapper
except ImportError as e:
    raise ImportError(
        'DatabasePushNotificationConfigStore requires SQLAlchemy and a database driver. '
        'Install with one of: '
        "'pip install a2a-sdk[postgresql]', "
        "'pip install a2a-sdk[mysql]', "
        "'pip install a2a-sdk[sqlite]', "
        "or 'pip install a2a-sdk[sql]'"
    ) from e

from collections.abc import Callable

from a2a.compat.v0_3.model_conversions import (
    compat_push_notification_config_model_to_core,
)
from a2a.server.context import ServerCallContext
from a2a.server.models import (
    Base,
    PushNotificationConfigModel,
    create_push_notification_config_model,
)
from a2a.server.owner_resolver import OwnerResolver, resolve_user_scope
from a2a.server.tasks.push_notification_config_store import (
    PushNotificationConfigStore,
)
from a2a.types.a2a_pb2 import TaskPushNotificationConfig


if TYPE_CHECKING:
    from cryptography.fernet import Fernet

logger = logging.getLogger(__name__)


class DatabasePushNotificationConfigStore(PushNotificationConfigStore):
    """SQLAlchemy-based implementation of PushNotificationConfigStore.

    Stores push notification configurations in a database supported by SQLAlchemy.
    """

    engine: AsyncEngine
    async_session_maker: async_sessionmaker[AsyncSession]
    create_table: bool
    _initialized: bool
    config_model: type[PushNotificationConfigModel]
    _fernet: 'Fernet | None'
    owner_resolver: OwnerResolver
    core_to_model_conversion: (
        Callable[
            [str, TaskPushNotificationConfig, str, 'Fernet | None'],
            PushNotificationConfigModel,
        ]
        | None
    )
    model_to_core_conversion: (
        Callable[[PushNotificationConfigModel], TaskPushNotificationConfig]
        | None
    )

    def __init__(  # noqa: PLR0913
        self,
        engine: AsyncEngine,
        create_table: bool = True,
        table_name: str = 'push_notification_configs',
        encryption_key: str | bytes | None = None,
        owner_resolver: OwnerResolver = resolve_user_scope,
        core_to_model_conversion: Callable[
            [str, TaskPushNotificationConfig, str, 'Fernet | None'],
            PushNotificationConfigModel,
        ]
        | None = None,
        model_to_core_conversion: Callable[
            [PushNotificationConfigModel], TaskPushNotificationConfig
        ]
        | None = None,
    ) -> None:
        """Initializes the DatabasePushNotificationConfigStore.

        Args:
            engine: An existing SQLAlchemy AsyncEngine to be used by the store.
            create_table: If true, create the table on initialization.
            table_name: Name of the database table. Defaults to 'push_notification_configs'.
            encryption_key: A key for encrypting sensitive configuration data.
                If provided, `config_data` will be encrypted in the database.
                The key must be a URL-safe base64-encoded 32-byte key.
            owner_resolver: Function to resolve the owner from the context.
            core_to_model_conversion: Optional function to convert a TaskPushNotificationConfig to a TaskPushNotificationConfigModel.
            model_to_core_conversion: Optional function to convert a TaskPushNotificationConfigModel to a TaskPushNotificationConfig.
        """
        logger.debug(
            'Initializing DatabasePushNotificationConfigStore with existing engine, table: %s',
            table_name,
        )
        self.engine = engine
        self.async_session_maker = async_sessionmaker(
            self.engine, expire_on_commit=False
        )
        self.create_table = create_table
        self._initialized = False
        self.owner_resolver = owner_resolver
        self.config_model = (
            PushNotificationConfigModel
            if table_name == 'push_notification_configs'
            else create_push_notification_config_model(table_name)
        )
        self._fernet = None
        self.core_to_model_conversion = core_to_model_conversion
        self.model_to_core_conversion = model_to_core_conversion

        if encryption_key:
            try:
                from cryptography.fernet import (
                    Fernet,
                )
            except ImportError as e:
                raise ImportError(
                    "DatabasePushNotificationConfigStore with encryption requires the 'cryptography' "
                    'library. Install with: '
                    "'pip install a2a-sdk[encryption]'"
                ) from e

            if isinstance(encryption_key, str):
                encryption_key = encryption_key.encode('utf-8')
            self._fernet = Fernet(encryption_key)
            logger.debug(
                'Encryption enabled for push notification config store.'
            )

    async def initialize(self) -> None:
        """Initialize the database and create the table if needed."""
        if self._initialized:
            return

        logger.debug(
            'Initializing database schema for push notification configs...'
        )
        if self.create_table:
            async with self.engine.begin() as conn:
                mapper = class_mapper(self.config_model)
                tables_to_create = [
                    table for table in mapper.tables if isinstance(table, Table)
                ]
                await conn.run_sync(
                    Base.metadata.create_all, tables=tables_to_create
                )
        self._initialized = True
        logger.debug(
            'Database schema for push notification configs initialized.'
        )

    async def _ensure_initialized(self) -> None:
        """Ensure the database connection is initialized."""
        if not self._initialized:
            await self.initialize()

    def _to_orm(
        self, task_id: str, config: TaskPushNotificationConfig, owner: str
    ) -> PushNotificationConfigModel:
        """Maps a TaskPushNotificationConfig proto to a SQLAlchemy model instance.

        The config data is serialized to JSON bytes, and encrypted if a key is configured.
        """
        if self.core_to_model_conversion:
            return self.core_to_model_conversion(
                task_id, config, owner, self._fernet
            )

        json_payload = MessageToJson(config).encode('utf-8')

        if self._fernet:
            data_to_store = self._fernet.encrypt(json_payload)
        else:
            data_to_store = json_payload

        return self.config_model(
            task_id=task_id,
            config_id=config.id,
            owner=owner,
            config_data=data_to_store,
            protocol_version='1.0',
        )

    def _from_orm(
        self, model_instance: PushNotificationConfigModel
    ) -> TaskPushNotificationConfig:
        """Maps a SQLAlchemy model instance to a TaskPushNotificationConfig proto.

        Handles decryption if a key is configured, with a fallback to plain JSON.
        """
        if self.model_to_core_conversion:
            return self.model_to_core_conversion(model_instance)

        payload = model_instance.config_data

        if self._fernet:
            from cryptography.fernet import (
                InvalidToken,
            )

            try:
                decrypted_payload = self._fernet.decrypt(payload)
                return self._parse_config(
                    decrypted_payload.decode('utf-8'),
                    model_instance.task_id,
                    model_instance.protocol_version,
                )
            except Exception as e:
                if isinstance(e, InvalidToken):
                    # Decryption failed. This could be because the data is not encrypted.
                    # We'll log a warning and try to parse it as plain JSON as a fallback.
                    logger.warning(
                        'Failed to decrypt push notification config for task %s, config %s. '
                        'Attempting to parse as unencrypted JSON. '
                        'This may indicate an incorrect encryption key or unencrypted data in the database.',
                        model_instance.task_id,
                        model_instance.config_id,
                    )
                    # Fall through to the unencrypted parsing logic below.
                else:
                    logger.exception(
                        'Failed to parse decrypted push notification config for task %s, config %s. '
                        'Data is corrupted or not valid JSON after decryption.',
                        model_instance.task_id,
                        model_instance.config_id,
                    )
                    raise ValueError(  # noqa: TRY004
                        'Failed to parse decrypted push notification config data'
                    ) from e

        # Try to parse as plain JSON.
        try:
            payload_str = (
                payload.decode('utf-8')
                if isinstance(payload, bytes)
                else payload
            )
            return self._parse_config(
                payload_str,
                model_instance.task_id,
                model_instance.protocol_version,
            )

        except Exception as e:
            if self._fernet:
                logger.exception(
                    'Failed to parse push notification config for task %s, config %s. '
                    'Decryption failed and the data is not valid JSON. '
                    'This likely indicates the data is corrupted or encrypted with a different key.',
                    model_instance.task_id,
                    model_instance.config_id,
                )
            else:
                # if no key is configured and the payload is not valid JSON.
                logger.exception(
                    'Failed to parse push notification config for task %s, config %s. '
                    'Data is not valid JSON and no encryption key is configured.',
                    model_instance.task_id,
                    model_instance.config_id,
                )
            raise ValueError(
                'Failed to parse push notification config data. '
                'Data is not valid JSON, or it is encrypted with the wrong key.'
            ) from e

    async def set_info(
        self,
        task_id: str,
        notification_config: TaskPushNotificationConfig,
        context: ServerCallContext,
    ) -> None:
        """Sets or updates the push notification configuration for a task."""
        await self._ensure_initialized()
        owner = self.owner_resolver(context)

        # Create a copy of the config using proto CopyFrom
        config_to_save = TaskPushNotificationConfig()
        config_to_save.CopyFrom(notification_config)
        if not config_to_save.id:
            config_to_save.id = task_id

        db_config = self._to_orm(task_id, config_to_save, owner)
        async with self.async_session_maker.begin() as session:
            await session.merge(db_config)
            logger.debug(
                'Push notification config for task %s with config id %s for owner %s saved/updated.',
                task_id,
                config_to_save.id,
                owner,
            )

    async def _select_configs(
        self,
        *predicates: 'ColumnElement[bool]',
    ) -> list[TaskPushNotificationConfig]:
        """Loads configs matching the given predicates and decodes them."""
        await self._ensure_initialized()
        async with self.async_session_maker() as session:
            stmt = select(self.config_model).where(and_(*predicates))
            result = await session.execute(stmt)
            models = result.scalars().all()

            configs = []
            for model in models:
                try:
                    configs.append(self._from_orm(model))
                except ValueError:  # noqa: PERF203
                    logger.exception(
                        'Could not deserialize push notification config for task %s, config %s, owner %s',
                        model.task_id,
                        model.config_id,
                        model.owner,
                    )
            return configs

    async def get_info(
        self,
        task_id: str,
        context: ServerCallContext,
    ) -> list[TaskPushNotificationConfig]:
        """Retrieves all push notification configurations for a task, for the given owner.

        Used by the user-callable read endpoints.
        """
        owner = self.owner_resolver(context)
        return await self._select_configs(
            self.config_model.task_id == task_id,
            self.config_model.owner == owner,
        )

    async def get_info_for_dispatch(
        self,
        task_id: str,
    ) -> list[TaskPushNotificationConfig]:
        """Retrieves all push notification configurations for a task, across all owners.

        Used by the push-notification dispatch path.
        """
        return await self._select_configs(
            self.config_model.task_id == task_id,
        )

    async def delete_info(
        self,
        task_id: str,
        context: ServerCallContext,
        config_id: str | None = None,
    ) -> None:
        """Deletes push notification configurations for a task.

        If config_id is provided, only that specific configuration is deleted.
        If config_id is None, all configurations for the task for the owner are deleted.
        """
        await self._ensure_initialized()
        owner = self.owner_resolver(context)
        async with self.async_session_maker.begin() as session:
            stmt = delete(self.config_model).where(
                and_(
                    self.config_model.task_id == task_id,
                    self.config_model.owner == owner,
                )
            )
            if config_id is not None:
                stmt = stmt.where(self.config_model.config_id == config_id)

            result = await session.execute(stmt)

            if result.rowcount > 0:  # type: ignore[attr-defined]
                logger.info(
                    'Deleted %s push notification config(s) for task %s, owner %s.',
                    result.rowcount,  # type: ignore[attr-defined]
                    task_id,
                    owner,
                )
            else:
                logger.warning(
                    'Attempted to delete push notification config for task %s, owner %s with config_id: %s that does not exist.',
                    task_id,
                    owner,
                    config_id,
                )

    def _parse_config(
        self,
        json_payload: str,
        task_id: str | None = None,
        protocol_version: str | None = None,
    ) -> TaskPushNotificationConfig:
        """Parses a JSON payload into a TaskPushNotificationConfig proto.

        Args:
            json_payload: The JSON payload to parse.
            task_id: The unique identifier of the task. Only required for legacy
                (0.3) protocol versions.
            protocol_version: The protocol version used for serialization.
        """
        if protocol_version == '1.0':
            return Parse(json_payload, TaskPushNotificationConfig())

        return compat_push_notification_config_model_to_core(
            json_payload, task_id or ''
        )



"""Constants for well-known URIs used throughout the A2A Python SDK."""

from enum import Enum


AGENT_CARD_WELL_KNOWN_PATH = '/.well-known/agent-card.json'
DEFAULT_RPC_URL = '/'
DEFAULT_LIST_TASKS_PAGE_SIZE = 50
"""Default page size for the `tasks/list` method."""

MAX_LIST_TASKS_PAGE_SIZE = 100
"""Maximum page size for the `tasks/list` method."""


class TransportProtocol(str, Enum):
    """Transport protocol string constants."""

    JSONRPC = 'JSONRPC'
    HTTP_JSON = 'HTTP+JSON'
    GRPC = 'GRPC'


JSONRPC_PARSE_ERROR_CODE = -32700
VERSION_HEADER = 'A2A-Version'

PROTOCOL_VERSION_1_0 = '1.0'
PROTOCOL_VERSION_0_3 = '0.3'
PROTOCOL_VERSION_CURRENT = PROTOCOL_VERSION_1_0



from collections.abc import AsyncGenerator, AsyncIterator, Awaitable, Callable
from typing import Any

from a2a.client.client import (
    Client,
    ClientCallContext,
    ClientConfig,
)
from a2a.client.interceptors import (
    AfterArgs,
    BeforeArgs,
    ClientCallInterceptor,
)
from a2a.client.transports.base import ClientTransport
from a2a.types.a2a_pb2 import (
    AgentCard,
    CancelTaskRequest,
    DeleteTaskPushNotificationConfigRequest,
    GetExtendedAgentCardRequest,
    GetTaskPushNotificationConfigRequest,
    GetTaskRequest,
    ListTaskPushNotificationConfigsRequest,
    ListTaskPushNotificationConfigsResponse,
    ListTasksRequest,
    ListTasksResponse,
    SendMessageRequest,
    StreamResponse,
    SubscribeToTaskRequest,
    Task,
    TaskPushNotificationConfig,
)


class BaseClient(Client):
    """Base implementation of the A2A client, containing transport-independent logic."""

    def __init__(
        self,
        card: AgentCard,
        config: ClientConfig,
        transport: ClientTransport,
        interceptors: list[ClientCallInterceptor],
    ):
        super().__init__(interceptors)
        self._card = card
        self._config = config
        self._transport = transport
        self._interceptors = interceptors

    async def send_message(
        self,
        request: SendMessageRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> AsyncIterator[StreamResponse]:
        """Sends a message to the agent.

        This method handles both streaming and non-streaming (polling) interactions
        based on the client configuration and agent capabilities. It will yield
        events as they are received from the agent.

        Args:
            request: The message to send to the agent.
            context: Optional client call context.

        Yields:
            An async iterator of `StreamResponse`
        """
        self._apply_client_config(request)
        if not self._config.streaming or not self._card.capabilities.streaming:
            response = await self._execute_with_interceptors(
                input_data=request,
                method='send_message',
                context=context,
                transport_call=lambda req, ctx: self._transport.send_message(
                    req, context=ctx
                ),
            )

            # In non-streaming case we convert to a StreamResponse so that the
            # client always sees the same iterator.
            stream_response = StreamResponse()
            if response.HasField('task'):
                stream_response.task.CopyFrom(response.task)
            elif response.HasField('message'):
                stream_response.message.CopyFrom(response.message)
            else:
                raise ValueError('Response has neither task nor message')

            yield stream_response
            return

        async for event in self._execute_stream_with_interceptors(
            input_data=request,
            method='send_message_streaming',
            context=context,
            transport_call=lambda req, ctx: (
                self._transport.send_message_streaming(req, context=ctx)
            ),
        ):
            yield event

    def _apply_client_config(self, request: SendMessageRequest) -> None:
        request.configuration.return_immediately |= self._config.polling
        if (
            not request.configuration.HasField('task_push_notification_config')
            and self._config.push_notification_config
        ):
            request.configuration.task_push_notification_config.CopyFrom(
                self._config.push_notification_config
            )
        if (
            not request.configuration.accepted_output_modes
            and self._config.accepted_output_modes
        ):
            request.configuration.accepted_output_modes.extend(
                self._config.accepted_output_modes
            )

    async def _process_stream(
        self,
        stream: AsyncIterator[StreamResponse],
        before_args: BeforeArgs,
    ) -> AsyncGenerator[StreamResponse, None]:
        async for stream_response in stream:
            after_args = AfterArgs(
                result=stream_response,
                method=before_args.method,
                agent_card=self._card,
                context=before_args.context,
            )
            await self._intercept_after(after_args)
            yield after_args.result
            if after_args.result.HasField('message'):
                return

    async def get_task(
        self,
        request: GetTaskRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> Task:
        """Retrieves the current state and history of a specific task.

        Args:
            request: The `GetTaskRequest` object specifying the task ID.
            context: Optional client call context.

        Returns:
            A `Task` object representing the current state of the task.
        """
        return await self._execute_with_interceptors(
            input_data=request,
            method='get_task',
            context=context,
            transport_call=lambda req, ctx: self._transport.get_task(
                req, context=ctx
            ),
        )

    async def list_tasks(
        self,
        request: ListTasksRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> ListTasksResponse:
        """Retrieves tasks for an agent."""
        return await self._execute_with_interceptors(
            input_data=request,
            method='list_tasks',
            context=context,
            transport_call=lambda req, ctx: self._transport.list_tasks(
                req, context=ctx
            ),
        )

    async def cancel_task(
        self,
        request: CancelTaskRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> Task:
        """Requests the agent to cancel a specific task.

        Args:
            request: The `CancelTaskRequest` object specifying the task ID.
            context: Optional client call context.

        Returns:
            A `Task` object containing the updated task status.
        """
        return await self._execute_with_interceptors(
            input_data=request,
            method='cancel_task',
            context=context,
            transport_call=lambda req, ctx: self._transport.cancel_task(
                req, context=ctx
            ),
        )

    async def create_task_push_notification_config(
        self,
        request: TaskPushNotificationConfig,
        *,
        context: ClientCallContext | None = None,
    ) -> TaskPushNotificationConfig:
        """Sets or updates the push notification configuration for a specific task.

        Args:
            request: The `TaskPushNotificationConfig` object with the new configuration.
            context: Optional client call context.

        Returns:
            The created or updated `TaskPushNotificationConfig` object.
        """
        return await self._execute_with_interceptors(
            input_data=request,
            method='create_task_push_notification_config',
            context=context,
            transport_call=lambda req, ctx: (
                self._transport.create_task_push_notification_config(
                    req, context=ctx
                )
            ),
        )

    async def get_task_push_notification_config(
        self,
        request: GetTaskPushNotificationConfigRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> TaskPushNotificationConfig:
        """Retrieves the push notification configuration for a specific task.

        Args:
            request: The `GetTaskPushNotificationConfigParams` object specifying the task.
            context: Optional client call context.

        Returns:
            A `TaskPushNotificationConfig` object containing the configuration.
        """
        return await self._execute_with_interceptors(
            input_data=request,
            method='get_task_push_notification_config',
            context=context,
            transport_call=lambda req, ctx: (
                self._transport.get_task_push_notification_config(
                    req, context=ctx
                )
            ),
        )

    async def list_task_push_notification_configs(
        self,
        request: ListTaskPushNotificationConfigsRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> ListTaskPushNotificationConfigsResponse:
        """Lists push notification configurations for a specific task.

        Args:
            request: The `ListTaskPushNotificationConfigsRequest` object specifying the request.
            context: Optional client call context.

        Returns:
            A `ListTaskPushNotificationConfigsResponse` object.
        """
        return await self._execute_with_interceptors(
            input_data=request,
            method='list_task_push_notification_configs',
            context=context,
            transport_call=lambda req, ctx: (
                self._transport.list_task_push_notification_configs(
                    req, context=ctx
                )
            ),
        )

    async def delete_task_push_notification_config(
        self,
        request: DeleteTaskPushNotificationConfigRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> None:
        """Deletes the push notification configuration for a specific task.

        Args:
            request: The `DeleteTaskPushNotificationConfigRequest` object specifying the request.
            context: Optional client call context.
        """
        return await self._execute_with_interceptors(
            input_data=request,
            method='delete_task_push_notification_config',
            context=context,
            transport_call=lambda req, ctx: (
                self._transport.delete_task_push_notification_config(
                    req, context=ctx
                )
            ),
        )

    async def subscribe(
        self,
        request: SubscribeToTaskRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> AsyncIterator[StreamResponse]:
        """Resubscribes to a task's event stream.

        This is only available if both the client and server support streaming.

        Args:
            request: Parameters to identify the task to resubscribe to.
            context: Optional client call context.

        Yields:
            An async iterator of `StreamResponse` objects.

        Raises:
            NotImplementedError: If streaming is not supported by the client or server.
        """
        if not self._config.streaming or not self._card.capabilities.streaming:
            raise NotImplementedError(
                'client and/or server do not support resubscription.'
            )

        async for event in self._execute_stream_with_interceptors(
            input_data=request,
            method='subscribe',
            context=context,
            transport_call=lambda req, ctx: self._transport.subscribe(
                req, context=ctx
            ),
        ):
            yield event

    async def get_extended_agent_card(
        self,
        request: GetExtendedAgentCardRequest,
        *,
        context: ClientCallContext | None = None,
        signature_verifier: Callable[[AgentCard], None] | None = None,
    ) -> AgentCard:
        """Retrieves the agent's card.

        This will fetch the authenticated card if necessary and update the
        client's internal state with the new card.

        Args:
            request: The `GetExtendedAgentCardRequest` object specifying the request.
            context: Optional client call context.
            signature_verifier: A callable used to verify the agent card's signatures.

        Returns:
            The `AgentCard` for the agent.
        """
        card = await self._execute_with_interceptors(
            input_data=request,
            method='get_extended_agent_card',
            context=context,
            transport_call=lambda req, ctx: (
                self._transport.get_extended_agent_card(req, context=ctx)
            ),
        )
        if signature_verifier:
            signature_verifier(card)

        self._card = card
        return card

    async def close(self) -> None:
        """Closes the underlying transport."""
        await self._transport.close()

    async def _execute_with_interceptors(
        self,
        input_data: Any,
        method: str,
        context: ClientCallContext | None,
        transport_call: Callable[
            [Any, ClientCallContext | None], Awaitable[Any]
        ],
    ) -> Any:
        before_args = BeforeArgs(
            input=input_data,
            method=method,
            agent_card=self._card,
            context=context,
        )
        before_result = await self._intercept_before(before_args)

        if before_result is not None:
            early_after_args = AfterArgs(
                result=before_result['early_return'],
                method=method,
                agent_card=self._card,
                context=before_args.context,
            )
            await self._intercept_after(
                early_after_args,
                before_result['executed'],
            )
            return early_after_args.result

        result = await transport_call(before_args.input, before_args.context)

        after_args = AfterArgs(
            result=result,
            method=method,
            agent_card=self._card,
            context=before_args.context,
        )
        await self._intercept_after(after_args)

        return after_args.result

    async def _execute_stream_with_interceptors(
        self,
        input_data: Any,
        method: str,
        context: ClientCallContext | None,
        transport_call: Callable[
            [Any, ClientCallContext | None], AsyncIterator[StreamResponse]
        ],
    ) -> AsyncIterator[StreamResponse]:

        before_args = BeforeArgs(
            input=input_data,
            method=method,
            agent_card=self._card,
            context=context,
        )
        before_result = await self._intercept_before(before_args)

        if before_result is not None:
            after_args = AfterArgs(
                result=before_result['early_return'],
                method=method,
                agent_card=self._card,
                context=before_args.context,
            )
            await self._intercept_after(after_args, before_result['executed'])

            yield after_args.result
            return

        stream = transport_call(before_args.input, before_args.context)

        async for client_event in self._process_stream(stream, before_args):
            yield client_event

    async def _intercept_before(
        self,
        args: BeforeArgs,
    ) -> dict[str, Any] | None:
        if not self._interceptors:
            return None
        executed: list[ClientCallInterceptor] = []
        for interceptor in self._interceptors:
            await interceptor.before(args)
            executed.append(interceptor)
            if args.early_return:
                return {
                    'early_return': args.early_return,
                    'executed': executed,
                }
        return None

    async def _intercept_after(
        self,
        args: AfterArgs,
        interceptors: list[ClientCallInterceptor] | None = None,
    ) -> None:
        interceptors_to_use = (
            interceptors if interceptors is not None else self._interceptors
        )

        reversed_interceptors = list(reversed(interceptors_to_use))
        for interceptor in reversed_interceptors:
            await interceptor.after(args)
            if args.early_return:
                return



"""Helper functions for building A2A JSON-RPC responses."""

from typing import Any

from google.protobuf.json_format import MessageToDict
from google.protobuf.message import Message as ProtoMessage
from jsonrpc.jsonrpc2 import JSONRPC20Response

from a2a.compat.v0_3.conversions import to_compat_agent_card
from a2a.server.jsonrpc_models import (
    InternalError as JSONRPCInternalError,
)
from a2a.server.jsonrpc_models import (
    JSONRPCError,
)
from a2a.types.a2a_pb2 import (
    AgentCard,
    ListTasksResponse,
    Message,
    StreamResponse,
    Task,
    TaskArtifactUpdateEvent,
    TaskPushNotificationConfig,
    TaskStatusUpdateEvent,
)
from a2a.types.a2a_pb2 import (
    SendMessageResponse as SendMessageResponseProto,
)
from a2a.utils.errors import (
    JSON_RPC_ERROR_CODE_MAP,
    A2AError,
    ContentTypeNotSupportedError,
    ExtendedAgentCardNotConfiguredError,
    ExtensionSupportRequiredError,
    InternalError,
    InvalidAgentResponseError,
    InvalidParamsError,
    InvalidRequestError,
    MethodNotFoundError,
    PushNotificationNotSupportedError,
    TaskNotCancelableError,
    TaskNotFoundError,
    UnsupportedOperationError,
    VersionNotSupportedError,
)


EXCEPTION_MAP: dict[type[A2AError], type[JSONRPCError]] = {
    TaskNotFoundError: JSONRPCError,
    TaskNotCancelableError: JSONRPCError,
    PushNotificationNotSupportedError: JSONRPCError,
    UnsupportedOperationError: JSONRPCError,
    ContentTypeNotSupportedError: JSONRPCError,
    InvalidAgentResponseError: JSONRPCError,
    ExtendedAgentCardNotConfiguredError: JSONRPCError,
    InvalidParamsError: JSONRPCError,
    InvalidRequestError: JSONRPCError,
    MethodNotFoundError: JSONRPCError,
    InternalError: JSONRPCInternalError,
    ExtensionSupportRequiredError: JSONRPCError,
    VersionNotSupportedError: JSONRPCError,
}


# Tuple of all A2AError types for isinstance checks
_A2A_ERROR_TYPES: tuple[type, ...] = (A2AError,)


# Result types for handler responses
EventTypes = (
    Task
    | Message
    | TaskArtifactUpdateEvent
    | TaskStatusUpdateEvent
    | TaskPushNotificationConfig
    | StreamResponse
    | SendMessageResponseProto
    | A2AError
    | JSONRPCError
    | list[TaskPushNotificationConfig]
    | ListTasksResponse
)
"""Type alias for possible event types produced by handlers."""


def agent_card_to_dict(card: AgentCard) -> dict[str, Any]:
    """Convert AgentCard to dict and inject backward compatibility fields."""
    result = MessageToDict(card)

    try:
        compat_card = to_compat_agent_card(card)
        compat_dict = compat_card.model_dump(exclude_none=True)
    except VersionNotSupportedError:
        compat_dict = {}

    # Do not include supportsAuthenticatedExtendedCard if false
    if not compat_dict.get('supportsAuthenticatedExtendedCard'):
        compat_dict.pop('supportsAuthenticatedExtendedCard', None)

    def merge(dict1: dict[str, Any], dict2: dict[str, Any]) -> dict[str, Any]:
        for k, v in dict2.items():
            if k not in dict1:
                dict1[k] = v
            elif isinstance(v, dict) and isinstance(dict1[k], dict):
                merge(dict1[k], v)
            elif isinstance(v, list) and isinstance(dict1[k], list):
                for i in range(min(len(dict1[k]), len(v))):
                    if isinstance(dict1[k][i], dict) and isinstance(v[i], dict):
                        merge(dict1[k][i], v[i])
        return dict1

    return merge(result, compat_dict)


def build_error_response(
    request_id: str | int | None,
    error: A2AError | JSONRPCError,
) -> dict[str, Any]:
    """Build a JSON-RPC error response dict.

    Args:
        request_id: The ID of the request that caused the error.
        error: The A2AError or JSONRPCError object.

    Returns:
        A dict representing the JSON-RPC error response.
    """
    jsonrpc_error: JSONRPCError
    if isinstance(error, JSONRPCError):
        jsonrpc_error = error
    elif isinstance(error, A2AError):
        error_type = type(error)
        model_class = EXCEPTION_MAP.get(error_type, JSONRPCInternalError)
        code = JSON_RPC_ERROR_CODE_MAP.get(error_type, -32603)
        jsonrpc_error = model_class(
            code=code,
            message=str(error),
            data=error.data,
        )
    else:
        jsonrpc_error = JSONRPCInternalError(message=str(error))

    error_dict = jsonrpc_error.model_dump(exclude_none=True)
    return JSONRPC20Response(error=error_dict, _id=request_id).data


def prepare_response_object(
    request_id: str | int | None,
    response: EventTypes,
    success_response_types: tuple[type, ...],
) -> dict[str, Any]:
    """Build a JSON-RPC response dict from handler output.

    Based on the type of the `response` object received from the handler,
    it constructs either a success response or an error response.

    Args:
        request_id: The ID of the request.
        response: The object received from the request handler.
        success_response_types: A tuple of expected types for a successful result.

    Returns:
        A dict representing the JSON-RPC response (success or error).
    """
    if isinstance(response, success_response_types):
        # Convert proto message to dict for JSON serialization
        result: Any = response
        if isinstance(response, ProtoMessage):
            result = MessageToDict(response, preserving_proto_field_name=False)
        return JSONRPC20Response(result=result, _id=request_id).data

    if isinstance(response, A2AError | JSONRPCError):
        return build_error_response(request_id, response)

    # If response is not an expected success type and not an error,
    # it's an invalid type of response from the agent for this method.
    error = InvalidAgentResponseError(
        message='Agent returned invalid type response for this method'
    )
    return build_error_response(request_id, error)



import logging

from collections.abc import AsyncGenerator
from typing import Any, NoReturn
from uuid import uuid4

import httpx

from google.protobuf import json_format
from jsonrpc.jsonrpc2 import JSONRPC20Request, JSONRPC20Response

from a2a.client.client import ClientCallContext
from a2a.client.errors import A2AClientError
from a2a.client.transports.base import ClientTransport
from a2a.client.transports.http_helpers import (
    get_http_args,
    send_http_request,
    send_http_stream_request,
)
from a2a.types.a2a_pb2 import (
    AgentCard,
    CancelTaskRequest,
    DeleteTaskPushNotificationConfigRequest,
    GetExtendedAgentCardRequest,
    GetTaskPushNotificationConfigRequest,
    GetTaskRequest,
    ListTaskPushNotificationConfigsRequest,
    ListTaskPushNotificationConfigsResponse,
    ListTasksRequest,
    ListTasksResponse,
    SendMessageRequest,
    SendMessageResponse,
    StreamResponse,
    SubscribeToTaskRequest,
    Task,
    TaskPushNotificationConfig,
)
from a2a.utils.errors import JSON_RPC_ERROR_CODE_MAP
from a2a.utils.telemetry import SpanKind, trace_class


logger = logging.getLogger(__name__)

_JSON_RPC_ERROR_CODE_TO_A2A_ERROR = {
    code: error_type for error_type, code in JSON_RPC_ERROR_CODE_MAP.items()
}


@trace_class(kind=SpanKind.CLIENT)
class JsonRpcTransport(ClientTransport):
    """A JSON-RPC transport for the A2A client."""

    def __init__(
        self,
        httpx_client: httpx.AsyncClient,
        agent_card: AgentCard,
        url: str,
    ):
        """Initializes the JsonRpcTransport."""
        self.url = url
        self.httpx_client = httpx_client
        self.agent_card = agent_card

    async def send_message(
        self,
        request: SendMessageRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> SendMessageResponse:
        """Sends a non-streaming message request to the agent."""
        rpc_request = JSONRPC20Request(
            method='SendMessage',
            params=json_format.MessageToDict(request),
            _id=str(uuid4()),
        )
        response_data = await self._send_request(
            dict(rpc_request.data), context
        )
        json_rpc_response = JSONRPC20Response(**response_data)
        if json_rpc_response.error:
            raise self._create_jsonrpc_error(json_rpc_response.error)
        response: SendMessageResponse = json_format.ParseDict(
            json_rpc_response.result, SendMessageResponse()
        )
        return response

    async def send_message_streaming(
        self,
        request: SendMessageRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> AsyncGenerator[StreamResponse]:
        """Sends a streaming message request to the agent and yields responses as they arrive."""
        rpc_request = JSONRPC20Request(
            method='SendStreamingMessage',
            params=json_format.MessageToDict(request),
            _id=str(uuid4()),
        )
        async for event in self._send_stream_request(
            dict(rpc_request.data),
            context,
        ):
            yield event

    async def get_task(
        self,
        request: GetTaskRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> Task:
        """Retrieves the current state and history of a specific task."""
        rpc_request = JSONRPC20Request(
            method='GetTask',
            params=json_format.MessageToDict(request),
            _id=str(uuid4()),
        )
        response_data = await self._send_request(
            dict(rpc_request.data), context
        )
        json_rpc_response = JSONRPC20Response(**response_data)
        if json_rpc_response.error:
            raise self._create_jsonrpc_error(json_rpc_response.error)
        response: Task = json_format.ParseDict(json_rpc_response.result, Task())
        return response

    async def list_tasks(
        self,
        request: ListTasksRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> ListTasksResponse:
        """Retrieves tasks for an agent."""
        rpc_request = JSONRPC20Request(
            method='ListTasks',
            params=json_format.MessageToDict(request),
            _id=str(uuid4()),
        )
        response_data = await self._send_request(
            dict(rpc_request.data), context
        )
        json_rpc_response = JSONRPC20Response(**response_data)
        if json_rpc_response.error:
            raise self._create_jsonrpc_error(json_rpc_response.error)
        response: ListTasksResponse = json_format.ParseDict(
            json_rpc_response.result, ListTasksResponse()
        )
        return response

    async def cancel_task(
        self,
        request: CancelTaskRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> Task:
        """Requests the agent to cancel a specific task."""
        rpc_request = JSONRPC20Request(
            method='CancelTask',
            params=json_format.MessageToDict(request),
            _id=str(uuid4()),
        )
        response_data = await self._send_request(
            dict(rpc_request.data), context
        )
        json_rpc_response = JSONRPC20Response(**response_data)
        if json_rpc_response.error:
            raise self._create_jsonrpc_error(json_rpc_response.error)
        response: Task = json_format.ParseDict(json_rpc_response.result, Task())
        return response

    async def create_task_push_notification_config(
        self,
        request: TaskPushNotificationConfig,
        *,
        context: ClientCallContext | None = None,
    ) -> TaskPushNotificationConfig:
        """Sets or updates the push notification configuration for a specific task."""
        rpc_request = JSONRPC20Request(
            method='CreateTaskPushNotificationConfig',
            params=json_format.MessageToDict(request),
            _id=str(uuid4()),
        )
        response_data = await self._send_request(
            dict(rpc_request.data), context
        )
        json_rpc_response = JSONRPC20Response(**response_data)
        if json_rpc_response.error:
            raise self._create_jsonrpc_error(json_rpc_response.error)
        response: TaskPushNotificationConfig = json_format.ParseDict(
            json_rpc_response.result, TaskPushNotificationConfig()
        )
        return response

    async def get_task_push_notification_config(
        self,
        request: GetTaskPushNotificationConfigRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> TaskPushNotificationConfig:
        """Retrieves the push notification configuration for a specific task."""
        rpc_request = JSONRPC20Request(
            method='GetTaskPushNotificationConfig',
            params=json_format.MessageToDict(request),
            _id=str(uuid4()),
        )
        response_data = await self._send_request(
            dict(rpc_request.data), context
        )
        json_rpc_response = JSONRPC20Response(**response_data)
        if json_rpc_response.error:
            raise self._create_jsonrpc_error(json_rpc_response.error)
        response: TaskPushNotificationConfig = json_format.ParseDict(
            json_rpc_response.result, TaskPushNotificationConfig()
        )
        return response

    async def list_task_push_notification_configs(
        self,
        request: ListTaskPushNotificationConfigsRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> ListTaskPushNotificationConfigsResponse:
        """Lists push notification configurations for a specific task."""
        rpc_request = JSONRPC20Request(
            method='ListTaskPushNotificationConfigs',
            params=json_format.MessageToDict(request),
            _id=str(uuid4()),
        )
        response_data = await self._send_request(
            dict(rpc_request.data), context
        )
        json_rpc_response = JSONRPC20Response(**response_data)
        if json_rpc_response.error:
            raise self._create_jsonrpc_error(json_rpc_response.error)
        response: ListTaskPushNotificationConfigsResponse = (
            json_format.ParseDict(
                json_rpc_response.result,
                ListTaskPushNotificationConfigsResponse(),
            )
        )
        return response

    async def delete_task_push_notification_config(
        self,
        request: DeleteTaskPushNotificationConfigRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> None:
        """Deletes the push notification configuration for a specific task."""
        rpc_request = JSONRPC20Request(
            method='DeleteTaskPushNotificationConfig',
            params=json_format.MessageToDict(request),
            _id=str(uuid4()),
        )
        response_data = await self._send_request(
            dict(rpc_request.data), context
        )
        json_rpc_response = JSONRPC20Response(**response_data)
        if json_rpc_response.error:
            raise self._create_jsonrpc_error(json_rpc_response.error)

    async def subscribe(
        self,
        request: SubscribeToTaskRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> AsyncGenerator[StreamResponse]:
        """Reconnects to get task updates."""
        rpc_request = JSONRPC20Request(
            method='SubscribeToTask',
            params=json_format.MessageToDict(request),
            _id=str(uuid4()),
        )
        async for event in self._send_stream_request(
            dict(rpc_request.data),
            context,
        ):
            yield event

    async def get_extended_agent_card(
        self,
        request: GetExtendedAgentCardRequest,
        *,
        context: ClientCallContext | None = None,
    ) -> AgentCard:
        """Retrieves the agent's card."""
        card = self.agent_card
        if not card.capabilities.extended_agent_card:
            return card

        rpc_request = JSONRPC20Request(
            method='GetExtendedAgentCard',
            params=json_format.MessageToDict(request),
            _id=str(uuid4()),
        )
        response_data = await self._send_request(
            dict(rpc_request.data),
            context,
        )
        json_rpc_response = JSONRPC20Response(**response_data)
        if json_rpc_response.error:
            raise self._create_jsonrpc_error(json_rpc_response.error)
        # Validate type of the response
        if not isinstance(json_rpc_response.result, dict):
            raise A2AClientError(
                f'Invalid response type: {type(json_rpc_response.result)}'
            )
        response: AgentCard = json_format.ParseDict(
            json_rpc_response.result, AgentCard()
        )

        return response

    async def close(self) -> None:
        """Closes the httpx client."""
        await self.httpx_client.aclose()

    def _create_jsonrpc_error(self, error_dict: dict[str, Any]) -> Exception:
        """Creates the appropriate A2AError from a JSON-RPC error dictionary."""
        code = error_dict.get('code')
        message = error_dict.get('message', str(error_dict))
        data = error_dict.get('data')

        if isinstance(code, int) and code in _JSON_RPC_ERROR_CODE_TO_A2A_ERROR:
            return _JSON_RPC_ERROR_CODE_TO_A2A_ERROR[code](message, data=data)

        # Fallback to general A2AClientError
        return A2AClientError(f'JSON-RPC Error {code}: {message}')

    async def _send_request(
        self,
        payload: dict[str, Any],
        context: ClientCallContext | None = None,
    ) -> dict[str, Any]:
        http_kwargs = get_http_args(context)

        request = self.httpx_client.build_request(
            'POST', self.url, json=payload, **(http_kwargs or {})
        )
        return await send_http_request(self.httpx_client, request)

    async def _send_stream_request(
        self,
        rpc_request_payload: dict[str, Any],
        context: ClientCallContext | None = None,
    ) -> AsyncGenerator[StreamResponse]:
        http_kwargs = get_http_args(context)

        async for sse_data in send_http_stream_request(
            self.httpx_client,
            'POST',
            self.url,
            None,
            self._handle_sse_error,
            json=rpc_request_payload,
            **http_kwargs,
        ):
            json_rpc_response = JSONRPC20Response.from_json(sse_data)
            if json_rpc_response.error:
                raise self._create_jsonrpc_error(json_rpc_response.error)
            response: StreamResponse = json_format.ParseDict(
                json_rpc_response.result, StreamResponse()
            )
            yield response

    def _handle_sse_error(self, sse_data: str) -> NoReturn:
        """Handles SSE error events by parsing JSON-RPC error payload and raising the appropriate domain error."""
        json_rpc_response = JSONRPC20Response.from_json(sse_data)
        if json_rpc_response.error:
            raise self._create_jsonrpc_error(json_rpc_response.error)
        raise A2AClientError(f'SSE stream error: {sse_data}')
